"""Pronunciation dictionaries for use in alignment and transcription"""

from __future__ import annotations

import logging
import math
import os
import re
import subprocess
import sys
from collections import Counter, defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Dict,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

import yaml

from .config.base_config import (
    DEFAULT_CLITIC_MARKERS,
    DEFAULT_COMPOUND_MARKERS,
    DEFAULT_DIGRAPHS,
    DEFAULT_PUNCTUATION,
    DEFAULT_STRIP_DIACRITICS,
)
from .exceptions import DictionaryError, DictionaryFileError, DictionaryPathError
from .utils import get_available_dictionaries, get_dictionary_path, thirdparty_binary

if TYPE_CHECKING:
    IpaType = Optional[List[str]]
    PunctuationType = Optional[str]
    from logging import Logger

    from .corpus.classes import Speaker

DictionaryEntryType = List[Dict[str, Union[Tuple[str], float, None, int]]]
ReversedMappingType = Dict[int, str]
WordsType = Dict[str, DictionaryEntryType]
MappingType = Dict[str, int]
MultiSpeakerMappingType = Dict[str, str]

__all__ = [
    "compile_graphemes",
    "sanitize",
    "check_format",
    "check_bracketed",
    "parse_ipa",
    "DictionaryData",
    "Dictionary",
    "MultispeakerDictionary",
]


def compile_graphemes(graphemes: Set[str]) -> re.Pattern:
    """
    Compiles the list of graphemes into a regular expression pattern.

    Parameters
    ----------
    graphemes: Set[str]
        Set of characters to treat as orthographic text

    Returns
    -------
    re.Pattern
        Compiled pattern that matches all graphemes
    """
    base = r"^\W*([{}]+)\W*"
    string = re.escape("".join(graphemes))
    try:
        return re.compile(base.format(string))
    except Exception:
        print(graphemes)
        raise


def check_bracketed(word: str, brackets: Optional[List[Tuple[str, str]]] = None) -> bool:
    """
    Checks whether a given string is surrounded by brackets.

    Parameters
    ----------
    word : str
        Text to check for final brackets
    brackets: List[Tuple[str, str]]], optional
        Brackets to check, defaults to [('[', ']'), ('{', '}'), ('<', '>'), ('(', ')')]

    Returns
    -------
    bool
        True if the word is fully bracketed, false otherwise
    """
    if brackets is None:
        brackets = [("[", "]"), ("{", "}"), ("<", ">"), ("(", ")")]
    for b in brackets:
        if word.startswith(b[0]) and word.endswith(b[-1]):
            return True
    return False


def sanitize(
    item: str, punctuation: Optional[str] = None, clitic_markers: Optional[str] = None
) -> str:
    """
    Sanitize an item according to punctuation and clitic markers

    Parameters
    ----------
    item: str
        Word to sanitize
    punctuation: str
        Characters to treat as punctuation
    clitic_markers: str
        Characters to treat as clitic markers, will be collapsed to the first marker

    Returns
    -------
    str
        Sanitized form
    """
    if punctuation is None:
        punctuation = DEFAULT_PUNCTUATION
    if clitic_markers is None:
        clitic_markers = DEFAULT_CLITIC_MARKERS
    for c in clitic_markers:
        item = item.replace(c, clitic_markers[0])
    if not item:
        return item
    if check_bracketed(item):
        return item
    sanitized = re.sub(rf"^[{punctuation}]+", "", item)
    sanitized = re.sub(rf"[{punctuation}]+$", "", sanitized)

    return sanitized


def check_format(path: str) -> Tuple[bool, bool]:
    """
    Check the pronunciation dictionary format

    Parameters
    ----------
    path: str
        Path of pronunciation dictionary

    Returns
    -------
    bool
        Flag for whether the dictionary has pronunciation probabilities
    bool
        Flag for whether the dictionary includes silence probabilities
    """
    count = 0
    pronunciation_probabilities = True
    silence_probabilities = True
    with open(path, "r", encoding="utf8") as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            line = line.split()
            _ = line.pop(0)  # word
            next_item = line.pop(0)
            if pronunciation_probabilities:
                try:
                    prob = float(next_item)
                    if prob > 1 or prob < 0:
                        raise ValueError
                except ValueError:
                    pronunciation_probabilities = False
            try:
                next_item = line.pop(0)
            except IndexError:
                silence_probabilities = False
            if silence_probabilities:
                try:
                    prob = float(next_item)
                    if prob > 1 or prob < 0:
                        raise ValueError
                except ValueError:
                    silence_probabilities = False
            count += 1
            if count > 10:
                break
    return pronunciation_probabilities, silence_probabilities


def parse_ipa(
    transcription: List[str], strip_diacritics: IpaType = None, digraphs: IpaType = None
) -> Tuple[str, ...]:
    """
    Parse a transcription in a multilingual IPA format (strips out diacritics and splits digraphs).

    Parameters
    ----------
    transcription: List[str]
        Transcription to parse
    strip_diacritics: List[str]
        List of diacritics to remove from characters in the transcription
    digraphs: List[str]
        List of digraphs to split up into separate characters

    Returns
    -------
    Tuple[str, ...]
        Parsed transcription
    """
    if strip_diacritics is None:
        strip_diacritics = DEFAULT_STRIP_DIACRITICS
    if digraphs is None:
        digraphs = DEFAULT_DIGRAPHS
    new_transcription = []
    for t in transcription:
        new_t = t
        for d in strip_diacritics:
            new_t = new_t.replace(d, "")
        if "g" in new_t:
            new_t = new_t.replace("g", "ɡ")

        found = False
        for digraph in digraphs:
            if re.match(r"^{}$".format(digraph), new_t):
                found = True
        if found:
            new_transcription.extend(new_t)
            continue
        new_transcription.append(new_t)
    return tuple(new_transcription)


class DictionaryData(NamedTuple):
    """
    Information required for parsing Kaldi-internal ids to text
    """

    silences: Set[str]
    multilingual_ipa: bool
    words_mapping: MappingType
    reversed_words_mapping: ReversedMappingType
    reversed_phone_mapping: ReversedMappingType
    punctuation: PunctuationType
    clitic_set: Set[str]
    clitic_markers: PunctuationType
    compound_markers: PunctuationType
    strip_diacritics: IpaType
    oov_int: int
    oov_code: str
    words: WordsType


class Dictionary:
    """
    Class containing information about a pronunciation dictionary

    Parameters
    ----------
    input_path : str
        Path to an input pronunciation dictionary
    output_directory : str
        Path to a directory to store files for Kaldi
    oov_code : str, optional
        What to label words not in the dictionary, defaults to ``'<unk>'``
    position_dependent_phones : bool, optional
        Specifies whether phones should be represented as dependent on their
        position in the word (beginning, middle or end), defaults to True
    num_sil_states : int, optional
        Number of states to use for silence phones, defaults to 5
    num_nonsil_states : int, optional
        Number of states to use for non-silence phones, defaults to 3
    shared_silence_phones : bool, optional
        Specify whether to share states across all silence phones, defaults
        to True
    sil_prob : float, optional
        Probability of optional silences following words, defaults to 0.5
    word_set : Collection[str], optional
        Word set to limit output files
    debug: bool, optional
        Flag for whether to perform debug steps and prevent intermediate cleanup
    logger: :class:`~logging.Logger`, optional
        Logger to output information to
    punctuation: str, optional
        Punctuation to use when parsing text
    clitic_markers: str, optional
        Clitic markers to use when parsing text
    compound_markers: str, optional
        Compound markers to use when parsing text
    multilingual_ipa: bool, optional
        Flag for multilingual IPA mode, defaults to False
    strip_diacritics: List[str], optional
        Diacritics to strip in multilingual IPA mode
    digraphs: List[str], optional
        Digraphs to split up in multilingual IPA mode
    """

    topo_template = "<State> {cur_state} <PdfClass> {cur_state} <Transition> {cur_state} 0.75 <Transition> {next_state} 0.25 </State>"
    topo_sil_template = "<State> {cur_state} <PdfClass> {cur_state} {transitions} </State>"
    topo_transition_template = "<Transition> {} {}"
    positions: List[str] = ["_B", "_E", "_I", "_S"]
    has_multiple = False

    def __init__(
        self,
        input_path: str,
        output_directory: str,
        oov_code: str = "<unk>",
        position_dependent_phones: bool = True,
        num_sil_states: int = 5,
        num_nonsil_states: int = 3,
        shared_silence_phones: bool = True,
        sil_prob: float = 0.5,
        word_set: Optional[Collection[str]] = None,
        debug: bool = False,
        logger: Optional[logging.Logger] = None,
        punctuation: PunctuationType = None,
        clitic_markers: PunctuationType = None,
        compound_markers: PunctuationType = None,
        multilingual_ipa: bool = False,
        strip_diacritics: IpaType = None,
        digraphs: IpaType = None,
    ):
        self.multilingual_ipa = multilingual_ipa
        self.strip_diacritics = DEFAULT_STRIP_DIACRITICS
        self.digraphs = DEFAULT_DIGRAPHS
        if strip_diacritics is not None:
            self.strip_diacritics = strip_diacritics
        if digraphs is not None:
            self.digraphs = digraphs
        self.punctuation = DEFAULT_PUNCTUATION
        self.clitic_markers = DEFAULT_CLITIC_MARKERS
        self.compound_markers = DEFAULT_COMPOUND_MARKERS
        if punctuation is not None:
            self.punctuation = punctuation
        if clitic_markers is not None:
            self.clitic_markers = clitic_markers
        if compound_markers is not None:
            self.compound_markers = compound_markers

        if not os.path.exists(input_path):
            raise (DictionaryPathError(input_path))
        if not os.path.isfile(input_path):
            raise (DictionaryFileError(input_path))
        self.input_path = input_path
        self.name = os.path.splitext(os.path.basename(input_path))[0]
        self.debug = debug
        self.output_directory = os.path.join(output_directory, self.name)
        os.makedirs(self.output_directory, exist_ok=True)
        self.log_file = os.path.join(self.output_directory, f"{self.name}.log")
        if logger is None:
            self.logger = logging.getLogger("dictionary_setup")
            self.logger.setLevel(logging.INFO)
            handler = logging.FileHandler(self.log_file, "w", "utf-8")
            handler.setFormatter = logging.Formatter("%(name)s %(message)s")
            self.logger.addHandler(handler)
            self.individual_logger = True
        else:
            self.logger = logger
            self.individual_logger = False
        self.num_sil_states = num_sil_states
        self.num_nonsil_states = num_nonsil_states
        self.shared_silence_phones = shared_silence_phones
        self.sil_prob = sil_prob
        self.oov_code = oov_code
        self.sil_code = "!sil"
        self.oovs_found = Counter()
        self.position_dependent_phones = position_dependent_phones

        self.words = {}
        self.nonsil_phones: Set[str] = set()
        self.sil_phones = {"sp", "spn", "sil"}
        self.optional_silence = "sp"
        self.nonoptional_silence = "sil"
        self.graphemes = set()
        self.max_disambiguation_symbol = 0
        self.disambiguation_symbols = set()
        self.all_words = defaultdict(list)
        self.clitic_set = set()
        self.specials_set = {self.oov_code, self.sil_code, "<eps>", "<s>", "</s>"}
        self.words[self.sil_code] = [{"pronunciation": ("sp",), "probability": 1}]
        self.words[self.oov_code] = [{"pronunciation": ("spn",), "probability": 1}]
        self.pronunciation_probabilities, self.silence_probabilities = check_format(input_path)
        progress = f'Parsing dictionary "{self.name}"'
        if self.pronunciation_probabilities:
            progress += " with pronunciation probabilities"
        else:
            progress += " without pronunciation probabilities"
        if self.silence_probabilities:
            progress += " with silence probabilities"
        else:
            progress += " without silence probabilities"
        self.logger.info(progress)
        with open(input_path, "r", encoding="utf8") as inf:
            for i, line in enumerate(inf):
                line = line.strip()
                if not line:
                    continue
                line = line.split()
                word = sanitize(line.pop(0).lower(), self.punctuation, self.clitic_markers)
                if not line:
                    raise DictionaryError(
                        f"Line {i} of {input_path} does not have a pronunciation."
                    )
                if word in ["!sil", oov_code]:
                    continue
                self.graphemes.update(word)
                prob = None
                if self.pronunciation_probabilities:
                    prob = float(line.pop(0))
                    if prob > 1 or prob < 0:
                        raise ValueError
                if self.silence_probabilities:
                    right_sil_prob = float(line.pop(0))
                    left_sil_prob = float(line.pop(0))
                    left_nonsil_prob = float(line.pop(0))
                else:
                    right_sil_prob = None
                    left_sil_prob = None
                    left_nonsil_prob = None
                if self.multilingual_ipa:
                    pron = parse_ipa(line, self.strip_diacritics, self.digraphs)
                else:
                    pron = tuple(line)
                pronunciation = {
                    "pronunciation": pron,
                    "probability": prob,
                    "disambiguation": None,
                    "right_sil_prob": right_sil_prob,
                    "left_sil_prob": left_sil_prob,
                    "left_nonsil_prob": left_nonsil_prob,
                }
                if self.multilingual_ipa:
                    pronunciation["original_pronunciation"] = tuple(line)
                if not any(x in self.sil_phones for x in pron):
                    self.nonsil_phones.update(pron)
                if word in self.words and pron in {x["pronunciation"] for x in self.words[word]}:
                    continue
                if word not in self.words:
                    self.words[word] = []
                self.words[word].append(pronunciation)
                # test whether a word is a clitic
                is_clitic = False
                for cm in self.clitic_markers:
                    if word.startswith(cm) or word.endswith(cm):
                        is_clitic = True
                if is_clitic:
                    self.clitic_set.add(word)
        if word_set is not None:
            word_set = {y for x in word_set for y in self._lookup(x)}
            word_set.add("!sil")
            word_set.add(self.oov_code)
        self.word_set = word_set
        if self.word_set is not None:
            self.word_set = self.word_set | self.clitic_set
        if not self.graphemes:
            raise DictionaryFileError(f"No words were found in the dictionary path {input_path}")
        self.word_pattern = compile_graphemes(self.graphemes)
        self.log_info()
        self.phone_mapping = {}
        self.words_mapping = {}

    def __hash__(self) -> Any:
        """Return the hash of a given dictionary"""
        return hash(self.input_path)

    @property
    def output_paths(self) -> Dict[str, str]:
        """
        Mapping of output directory for this dictionary
        """
        return {self.name: self.output_directory}

    @property
    def silences(self) -> Set[str]:
        """
        Set of symbols that correspond to silence
        """
        return {self.optional_silence, self.nonoptional_silence}

    def get_dictionary(self, speaker: Union[Speaker, str]) -> Dictionary:
        """
        Wrapper function to return this dictionary for any arbitrary speaker

        Parameters
        ----------
        speaker: Union[Speaker, str]
            Speaker to look up dictionary for

        Returns
        -------
        Dictionary
            This dictionary
        """
        return self

    def data(self, word_set: Optional[Collection[str]] = None) -> DictionaryData:
        """
        Generates a dictionary data for use in parsing utilities

        Parameters
        ----------
        word_set: Collection[str], optional
            Word set to limit data to

        Returns
        -------
        DictionaryData
            Data necessary for parsing text
        """

        def word_check(word):
            """Check whether a word should be included in the output"""
            if word in word_set:
                return True
            if word in self.clitic_set:
                return True
            if word in self.specials_set:
                return True
            return False

        if word_set:
            words_mapping = {k: v for k, v in self.words_mapping.items() if word_check(k)}
            reversed_word_mapping = {
                k: v for k, v in self.reversed_word_mapping.items() if word_check(v)
            }
            words = {k: v for k, v in self.words.items() if word_check(k)}
        else:
            words_mapping = self.words_mapping
            reversed_word_mapping = self.reversed_word_mapping
            words = self.words
        return DictionaryData(
            self.silences,
            self.multilingual_ipa,
            words_mapping,
            reversed_word_mapping,
            self.reversed_phone_mapping,
            self.punctuation,
            self.clitic_set,
            self.clitic_markers,
            self.compound_markers,
            self.strip_diacritics,
            self.oov_int,
            self.oov_code,
            words,
        )

    def cleanup_logger(self) -> None:
        """
        Clean up and detach logger from handles
        """
        if not self.individual_logger:
            return
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)

    def log_info(self) -> None:
        """
        Dump debugging information to the logger
        """
        self.logger.debug(f'"{self.name}" DICTIONARY INFORMATION')
        if self.pronunciation_probabilities:
            self.logger.debug("Has pronunciation probabilities")
        else:
            self.logger.debug("Has NO pronunciation probabilities")
        if self.silence_probabilities:
            self.logger.debug("Has silence probabilities")
        else:
            self.logger.debug("Has NO silence probabilities")

        self.logger.debug(f"Grapheme set: {', '.join(sorted(self.graphemes))}")
        self.logger.debug(f"Phone set: {', '.join(sorted(self.nonsil_phones))}")
        self.logger.debug(f"Punctuation: {self.punctuation}")
        self.logger.debug(f"Clitic markers: {self.clitic_markers}")
        self.logger.debug(f"Clitic set: {', '.join(sorted(self.clitic_set))}")
        if self.multilingual_ipa:
            self.logger.debug(f"Strip diacritics: {', '.join(sorted(self.strip_diacritics))}")
            self.logger.debug(f"Digraphs: {', '.join(sorted(self.digraphs))}")

    def set_word_set(self, word_set: List[str]) -> None:
        """
        Limit output to a subset of overall words

        Parameters
        ----------
        word_set: List[str]
            Word set to limit generated files to
        """
        word_set = {y for x in word_set for y in self._lookup(x)}
        word_set.add(self.sil_code)
        word_set.add(self.oov_code)
        self.word_set = word_set | self.clitic_set
        self.generate_mappings()

    @property
    def actual_words(self) -> Dict[str, DictionaryEntryType]:
        """
        Mapping of words to integer IDs without Kaldi-internal words
        """
        return {
            k: v
            for k, v in self.words.items()
            if k not in [self.sil_code, self.oov_code, "<eps>"] and len(v)
        }

    def split_clitics(self, item: str) -> List[str]:
        """
        Split a word into subwords based on clitic and compound markers

        Parameters
        ----------
        item: str
            Word to split up

        Returns
        -------
        List[str]
            List of subwords
        """
        if item in self.words:
            return [item]
        if any(x in item for x in self.compound_markers):
            s = re.split(rf"[{self.compound_markers}]", item)
            if any(x in item for x in self.clitic_markers):
                new_s = []
                for seg in s:
                    if any(x in seg for x in self.clitic_markers):
                        new_s.extend(self.split_clitics(seg))
                    else:
                        new_s.append(seg)
                s = new_s
            return s
        if any(
            x in item and not item.endswith(x) and not item.startswith(x)
            for x in self.clitic_markers
        ):
            initial, final = re.split(rf"[{self.clitic_markers}]", item, maxsplit=1)
            if any(x in final for x in self.clitic_markers):
                final = self.split_clitics(final)
            else:
                final = [final]
            for clitic in self.clitic_markers:
                if initial + clitic in self.clitic_set:
                    return [initial + clitic] + final
                elif clitic + final[0] in self.clitic_set:
                    final[0] = clitic + final[0]
                    return [initial] + final
        return [item]

    def __len__(self) -> int:
        """Return the number of pronunciations across all words"""
        return sum(len(x) for x in self.words.values())

    def exclude_for_alignment(self, word: str) -> bool:
        """
        Check for whether to exclude a word from alignment lexicons (if there is a word set in the dictionary,
        checks whether the given string is in the word set)

        Parameters
        ----------
        word: str
            Word to check

        Returns
        -------
        bool
            True if there is no word set on the dictionary, or if the word is in the given word set
        """
        if self.word_set is None:
            return False
        if word not in self.word_set and word not in self.clitic_set:
            return True
        return False

    def generate_mappings(self) -> None:
        """
        Generate phone and word mappings from text to integer IDs
        """
        if self.phone_mapping:
            return
        self.phone_mapping = {}
        i = 0
        self.phone_mapping["<eps>"] = i
        if self.position_dependent_phones:
            for p in self.positional_sil_phones:
                i += 1
                self.phone_mapping[p] = i
            for p in self.positional_nonsil_phones:
                i += 1
                self.phone_mapping[p] = i
        else:
            for p in sorted(self.sil_phones):
                i += 1
                self.phone_mapping[p] = i
            for p in sorted(self.nonsil_phones):
                i += 1
                self.phone_mapping[p] = i

        self.words_mapping = {}
        i = 0
        self.words_mapping["<eps>"] = i
        for w in sorted(self.words.keys()):
            if self.exclude_for_alignment(w):
                continue
            i += 1
            self.words_mapping[w] = i

        self.words_mapping["#0"] = i + 1
        self.words_mapping["<s>"] = i + 2
        self.words_mapping["</s>"] = i + 3
        self.oovs_found = Counter()
        self.add_disambiguation()

    def add_disambiguation(self) -> None:
        """
        Calculate disambiguation symbols for each pronunciation
        """
        subsequences = set()
        pronunciation_counts = defaultdict(int)

        for w, prons in self.words.items():
            if self.exclude_for_alignment(w):
                continue
            for p in prons:
                pronunciation_counts[p["pronunciation"]] += 1
                pron = p["pronunciation"][:-1]
                while pron:
                    subsequences.add(tuple(p))
                    pron = pron[:-1]
        last_used = defaultdict(int)
        for w, prons in sorted(self.words.items()):
            if self.exclude_for_alignment(w):
                continue
            for p in prons:
                if (
                    pronunciation_counts[p["pronunciation"]] == 1
                    and not p["pronunciation"] in subsequences
                ):
                    disambig = None
                else:
                    pron = p["pronunciation"]
                    last_used[pron] += 1
                    disambig = last_used[pron]
                p["disambiguation"] = disambig
        if last_used:
            self.max_disambiguation_symbol = max(last_used.values())
        else:
            self.max_disambiguation_symbol = 0
        self.disambiguation_symbols = set()
        i = max(self.phone_mapping.values())
        for x in range(self.max_disambiguation_symbol + 2):
            p = f"#{x}"
            self.disambiguation_symbols.add(p)
            i += 1
            self.phone_mapping[p] = i

    def create_utterance_fst(self, text: List[str], frequent_words: List[Tuple[str, int]]) -> str:
        """
        Create an FST for an utterance with frequent words as a unigram language model

        Parameters
        ----------
        text: List[str]
            Text of the utterance
        frequent_words: List[Tuple[str, int]]
            Frequent words to incorporate into the FST
        Returns
        -------
        str
            FST created from the utterance text and frequent words
        """
        num_words = len(text)
        word_probs = Counter(text)
        word_probs = {k: v / num_words for k, v in word_probs.items()}
        word_probs.update(frequent_words)
        fst_text = ""
        for k, v in word_probs.items():
            cost = -1 * math.log(v)
            w = self.to_int(k)[0]
            fst_text += f"0 0 {w} {w} {cost}\n"
        fst_text += f"0 {-1 * math.log(1 / num_words)}\n"
        return fst_text

    def to_int(self, item: str) -> List[int]:
        """
        Convert a given word into integer IDs

        Parameters
        ----------
        item: str
            Word to look up

        Returns
        -------
        List[int]
            List of integer IDs corresponding to each subword
        """
        if item == "":
            return []
        sanitized = self._lookup(item)
        text_int = []
        for item in sanitized:
            if not item:
                continue
            if item not in self.words_mapping:
                self.oovs_found.update([item])
                text_int.append(self.oov_int)
            else:
                text_int.append(self.words_mapping[item])
        return text_int

    def save_oovs_found(self, directory: str) -> None:
        """
        Save all out of vocabulary items to a file in the specified directory

        Parameters
        ----------
        directory : str
            Path to directory to save ``oovs_found.txt``
        """
        with open(os.path.join(directory, "oovs_found.txt"), "w", encoding="utf8") as f, open(
            os.path.join(directory, "oov_counts.txt"), "w", encoding="utf8"
        ) as cf:
            for oov in sorted(self.oovs_found.keys(), key=lambda x: (-self.oovs_found[x], x)):
                f.write(oov + "\n")
                cf.write(f"{oov}\t{self.oovs_found[oov]}\n")

    def _lookup(self, item: str) -> List[str]:
        """
        Look up a word and return the list of sub words if necessary taking into account clitic and compound markers

        Parameters
        ----------
        item: str
            Word to look up

        Returns
        -------
        List[str]
            List of subwords that are in the dictionary
        """
        if item in self.words:
            return [item]
        sanitized = sanitize(item, self.punctuation, self.clitic_markers)
        if sanitized in self.words:
            return [sanitized]
        split = self.split_clitics(sanitized)
        oov_count = sum(1 for x in split if x not in self.words)
        if oov_count < len(
            split
        ):  # Only returned split item if it gains us any transcribed speech
            return split
        return [sanitized]

    def check_word(self, item: str) -> bool:
        """
        Check whether a word is in the dictionary, takes into account sanitization and
        clitic and compound markers

        Parameters
        ----------
        item: str
            Word to check

        Returns
        -------
        bool
            True if the look up would not result in an OOV item
        """
        if item == "":
            return False
        if item in self.words:
            return True
        sanitized = sanitize(item, self.punctuation, self.clitic_markers)
        if sanitized in self.words:
            return True

        sanitized = self.split_clitics(sanitized)
        if all(s in self.words for s in sanitized):
            return True
        return False

    @property
    def reversed_word_mapping(self) -> ReversedMappingType:
        """
        A mapping of integer ids to words
        """
        mapping = {}
        for k, v in self.words_mapping.items():
            mapping[v] = k
        return mapping

    @property
    def reversed_phone_mapping(self) -> ReversedMappingType:
        """
        A mapping of integer ids to phones
        """
        mapping = {}
        for k, v in self.phone_mapping.items():
            mapping[v] = k
        return mapping

    @property
    def oov_int(self) -> int:
        """
        The integer id for out of vocabulary items
        """
        return self.words_mapping[self.oov_code]

    @property
    def positional_sil_phones(self) -> List[str]:
        """
        List of silence phones with positions
        """
        sil_phones = []
        for p in sorted(self.sil_phones):
            sil_phones.append(p)
            for pos in self.positions:
                sil_phones.append(p + pos)
        return sil_phones

    @property
    def positional_nonsil_phones(self) -> List[str]:
        """
        List of non-silence phones with positions
        """
        nonsil_phones = []
        for p in sorted(self.nonsil_phones):
            for pos in self.positions:
                nonsil_phones.append(p + pos)
        return nonsil_phones

    @property
    def optional_silence_csl(self) -> str:
        """
        Phone id of the optional silence phone
        """
        return str(self.phone_mapping[self.optional_silence])

    @property
    def silence_csl(self) -> str:
        """
        A colon-separated list (as a string) of silence phone ids
        """
        if self.position_dependent_phones:
            return ":".join(map(str, (self.phone_mapping[x] for x in self.positional_sil_phones)))
        else:
            return ":".join(map(str, (self.phone_mapping[x] for x in self.sil_phones)))

    @property
    def phones_dir(self) -> str:
        """
        Directory to store information Kaldi needs about phones
        """
        return os.path.join(self.output_directory, "phones")

    @property
    def phones(self) -> set:
        """
        The set of all phones (silence and non-silence)
        """
        return self.sil_phones | self.nonsil_phones

    @property
    def words_symbol_path(self) -> str:
        """
        Path of word to int mapping file for the dictionary
        """
        return os.path.join(self.output_directory, "words.txt")

    @property
    def disambig_path(self) -> str:
        """
        Path of disambiguated lexicon fst (L.fst)
        """
        return os.path.join(self.output_directory, "L_disambig.fst")

    def write(self, write_disambiguation: Optional[bool] = False) -> None:
        """
        Write the files necessary for Kaldi

        Parameters
        ----------
        write_disambiguation: bool, optional
            Flag for including disambiguation information
        """
        self.logger.info("Creating dictionary information...")
        os.makedirs(self.phones_dir, exist_ok=True)
        self.generate_mappings()
        self._write_graphemes()
        self._write_phone_map_file()
        self._write_phone_sets()
        self._write_phone_symbol_table()
        self._write_disambig()
        self._write_topo()
        self._write_word_boundaries()
        self._write_extra_questions()
        self._write_word_file()
        self._write_align_lexicon()
        self._write_fst_text(write_disambiguation=write_disambiguation)
        self._write_fst_binary(write_disambiguation=write_disambiguation)
        self.cleanup()

    def cleanup(self) -> None:
        """
        Clean up temporary files in the output directory
        """
        if not self.debug:
            if os.path.exists(os.path.join(self.output_directory, "temp.fst")):
                os.remove(os.path.join(self.output_directory, "temp.fst"))
            if os.path.exists(os.path.join(self.output_directory, "lexicon.text.fst")):
                os.remove(os.path.join(self.output_directory, "lexicon.text.fst"))

    def _write_graphemes(self) -> None:
        """
        Write graphemes to temporary directory
        """
        outfile = os.path.join(self.output_directory, "graphemes.txt")
        if os.path.exists(outfile):
            return
        with open(outfile, "w", encoding="utf8") as f:
            for char in sorted(self.graphemes):
                f.write(char + "\n")

    def export_lexicon(
        self,
        path: str,
        write_disambiguation: Optional[bool] = False,
        probability: Optional[bool] = False,
    ) -> None:
        """
        Export pronunciation dictionary to a text file

        Parameters
        ----------
        path: str
            Path to save dictionary
        write_disambiguation: bool, optional
            Flag for whether to include disambiguation information
        probability: bool, optional
            Flag for whether to include probabilities
        """
        with open(path, "w", encoding="utf8") as f:
            for w in sorted(self.words.keys()):
                for p in sorted(
                    self.words[w],
                    key=lambda x: (x["pronunciation"], x["probability"], x["disambiguation"]),
                ):
                    phones = " ".join(p["pronunciation"])
                    if write_disambiguation and p["disambiguation"] is not None:
                        phones += f" #{p['disambiguation']}"
                    if probability:
                        f.write(f"{w}\t{p['probability']}\t{phones}\n")
                    else:
                        f.write(f"{w}\t{phones}\n")

    def _write_phone_map_file(self) -> None:
        """
        Write the phone map to the temporary directory
        """
        outfile = os.path.join(self.output_directory, "phone_map.txt")
        if os.path.exists(outfile):
            return
        with open(outfile, "w", encoding="utf8") as f:
            for sp in self.sil_phones:
                if self.position_dependent_phones:
                    new_phones = [sp + x for x in ["", ""] + self.positions]
                else:
                    new_phones = [sp]
                f.write(" ".join(new_phones) + "\n")
            for nsp in self.nonsil_phones:
                if self.position_dependent_phones:
                    new_phones = [nsp + x for x in [""] + self.positions]
                else:
                    new_phones = [nsp]
                f.write(" ".join(new_phones) + "\n")

    def _write_phone_symbol_table(self) -> None:
        """
        Write the phone mapping to the temporary directory
        """
        outfile = os.path.join(self.output_directory, "phones.txt")
        if os.path.exists(outfile):
            return
        with open(outfile, "w", encoding="utf8") as f:
            for p, i in sorted(self.phone_mapping.items(), key=lambda x: x[1]):
                f.write(f"{p} {i}\n")

    def _write_word_boundaries(self) -> None:
        """
        Write the word boundaries file to the temporary directory
        """
        boundary_path = os.path.join(self.output_directory, "phones", "word_boundary.txt")
        boundary_int_path = os.path.join(self.output_directory, "phones", "word_boundary.int")
        if os.path.exists(boundary_path) and os.path.exists(boundary_int_path):
            return
        with open(boundary_path, "w", encoding="utf8") as f, open(
            boundary_int_path, "w", encoding="utf8"
        ) as intf:
            if self.position_dependent_phones:
                for p in sorted(self.phone_mapping.keys(), key=lambda x: self.phone_mapping[x]):
                    if p == "<eps>" or p.startswith("#"):
                        continue
                    cat = "nonword"
                    if p.endswith("_B"):
                        cat = "begin"
                    elif p.endswith("_S"):
                        cat = "singleton"
                    elif p.endswith("_I"):
                        cat = "internal"
                    elif p.endswith("_E"):
                        cat = "end"
                    f.write(" ".join([p, cat]) + "\n")
                    intf.write(" ".join([str(self.phone_mapping[p]), cat]) + "\n")

    def _write_word_file(self) -> None:
        """
        Write the word mapping to the temporary directory
        """
        words_path = os.path.join(self.output_directory, "words.txt")
        if os.path.exists(words_path):
            return
        if sys.platform == "win32":
            newline = ""
        else:
            newline = None
        with open(words_path, "w", encoding="utf8", newline=newline) as f:
            for w, i in sorted(self.words_mapping.items(), key=lambda x: x[1]):
                f.write(f"{w} {i}\n")

    def _write_align_lexicon(self) -> None:
        """
        Write the alignment lexicon text file to the temporary directory
        """
        path = os.path.join(self.phones_dir, "align_lexicon.int")
        if os.path.exists(path):
            return

        with open(path, "w", encoding="utf8") as f:
            for w, i in self.words_mapping.items():
                if self.exclude_for_alignment(w):
                    continue
                if w not in self.words:  # special characters
                    continue
                for pron in sorted(
                    self.words[w],
                    key=lambda x: (x["pronunciation"], x["probability"], x["disambiguation"]),
                ):

                    phones = list(pron["pronunciation"])
                    if self.position_dependent_phones:
                        if len(phones) == 1:
                            phones[0] += "_S"
                        else:
                            for j in range(len(phones)):
                                if j == 0:
                                    phones[j] += "_B"
                                elif j == len(phones) - 1:
                                    phones[j] += "_E"
                                else:
                                    phones[j] += "_I"
                    p = " ".join(str(self.phone_mapping[x]) for x in phones)
                    f.write(f"{i} {i} {p}\n".format(i=i, p=p))

    def _write_topo(self) -> None:
        """
        Write the topo file to the temporary directory
        """
        filepath = os.path.join(self.output_directory, "topo")
        if os.path.exists(filepath):
            return
        sil_transp = 1 / (self.num_sil_states - 1)
        initial_transition = [
            self.topo_transition_template.format(x, sil_transp)
            for x in range(self.num_sil_states - 1)
        ]
        middle_transition = [
            self.topo_transition_template.format(x, sil_transp)
            for x in range(1, self.num_sil_states)
        ]
        final_transition = [
            self.topo_transition_template.format(self.num_sil_states - 1, 0.75),
            self.topo_transition_template.format(self.num_sil_states, 0.25),
        ]
        with open(filepath, "w") as f:
            f.write("<Topology>\n")
            f.write("<TopologyEntry>\n")
            f.write("<ForPhones>\n")
            if self.position_dependent_phones:
                phones = self.positional_nonsil_phones
            else:
                phones = sorted(self.nonsil_phones)
            f.write(f"{' '.join(str(self.phone_mapping[x]) for x in phones)}\n")
            f.write("</ForPhones>\n")
            states = [
                self.topo_template.format(cur_state=x, next_state=x + 1)
                for x in range(self.num_nonsil_states)
            ]
            f.write("\n".join(states))
            f.write(f"\n<State> {self.num_nonsil_states} </State>\n")
            f.write("</TopologyEntry>\n")

            f.write("<TopologyEntry>\n")
            f.write("<ForPhones>\n")
            if self.position_dependent_phones:
                phones = self.positional_sil_phones
            else:
                phones = self.sil_phones
            f.write(f"{' '.join(str(self.phone_mapping[x]) for x in phones)}\n")
            f.write("</ForPhones>\n")
            states = []
            for i in range(self.num_sil_states):
                if i == 0:
                    transition = " ".join(initial_transition)
                elif i == self.num_sil_states - 1:
                    transition = " ".join(final_transition)
                else:
                    transition = " ".join(middle_transition)
                states.append(self.topo_sil_template.format(cur_state=i, transitions=transition))
            f.write("\n".join(states))
            f.write(f"\n<State> {self.num_sil_states} </State>\n")
            f.write("</TopologyEntry>\n")
            f.write("</Topology>\n")

    def _write_phone_sets(self) -> None:
        """
        Write phone symbol sets to the temporary directory
        """
        sharesplit = ["shared", "split"]
        if not self.shared_silence_phones:
            sil_sharesplit = ["not-shared", "not-split"]
        else:
            sil_sharesplit = sharesplit

        sets_file = os.path.join(self.output_directory, "phones", "sets.txt")
        roots_file = os.path.join(self.output_directory, "phones", "roots.txt")

        sets_int_file = os.path.join(self.output_directory, "phones", "sets.int")
        roots_int_file = os.path.join(self.output_directory, "phones", "roots.int")
        if (
            os.path.exists(sets_file)
            and os.path.exists(roots_file)
            and os.path.exists(sets_int_file)
            and os.path.exists(roots_int_file)
        ):
            return

        with open(sets_file, "w", encoding="utf8") as setf, open(
            roots_file, "w", encoding="utf8"
        ) as rootf, open(sets_int_file, "w", encoding="utf8") as setintf, open(
            roots_int_file, "w", encoding="utf8"
        ) as rootintf:

            # process silence phones
            for i, sp in enumerate(self.sil_phones):
                if self.position_dependent_phones:
                    mapped = [sp + x for x in [""] + self.positions]
                else:
                    mapped = [sp]
                setf.write(" ".join(mapped) + "\n")
                setintf.write(" ".join(map(str, (self.phone_mapping[x] for x in mapped))) + "\n")
                if i == 0:
                    line = sil_sharesplit + mapped
                    lineint = sil_sharesplit + [self.phone_mapping[x] for x in mapped]
                else:
                    line = sharesplit + mapped
                    lineint = sharesplit + [self.phone_mapping[x] for x in mapped]
                rootf.write(" ".join(line) + "\n")
                rootintf.write(" ".join(map(str, lineint)) + "\n")

            # process nonsilence phones
            for nsp in sorted(self.nonsil_phones):
                if self.position_dependent_phones:
                    mapped = [nsp + x for x in self.positions]
                else:
                    mapped = [nsp]
                setf.write(" ".join(mapped) + "\n")
                setintf.write(" ".join(map(str, (self.phone_mapping[x] for x in mapped))) + "\n")
                line = sharesplit + mapped
                lineint = sharesplit + [self.phone_mapping[x] for x in mapped]
                rootf.write(" ".join(line) + "\n")
                rootintf.write(" ".join(map(str, lineint)) + "\n")

    def _write_extra_questions(self) -> None:
        """
        Write extra questions symbols to the temporary directory
        """
        phone_extra = os.path.join(self.phones_dir, "extra_questions.txt")
        phone_extra_int = os.path.join(self.phones_dir, "extra_questions.int")
        if os.path.exists(phone_extra) and os.path.exists(phone_extra_int):
            return
        with open(phone_extra, "w", encoding="utf8") as outf, open(
            phone_extra_int, "w", encoding="utf8"
        ) as intf:
            if self.position_dependent_phones:
                sils = sorted(self.positional_sil_phones)
            else:
                sils = sorted(self.sil_phones)
            outf.write(" ".join(sils) + "\n")
            intf.write(" ".join(map(str, (self.phone_mapping[x] for x in sils))) + "\n")

            if self.position_dependent_phones:
                nonsils = sorted(self.positional_nonsil_phones)
            else:
                nonsils = sorted(self.nonsil_phones)
            outf.write(" ".join(nonsils) + "\n")
            intf.write(" ".join(map(str, (self.phone_mapping[x] for x in nonsils))) + "\n")
            if self.position_dependent_phones:
                for p in self.positions:
                    line = [x + p for x in sorted(self.nonsil_phones)]
                    outf.write(" ".join(line) + "\n")
                    intf.write(" ".join(map(str, (self.phone_mapping[x] for x in line))) + "\n")
                for p in [""] + self.positions:
                    line = [x + p for x in sorted(self.sil_phones)]
                    outf.write(" ".join(line) + "\n")
                    intf.write(" ".join(map(str, (self.phone_mapping[x] for x in line))) + "\n")

    def _write_disambig(self) -> None:
        """
        Write disambiguation symbols to the temporary directory
        """
        disambig = os.path.join(self.phones_dir, "disambiguation_symbols.txt")
        disambig_int = os.path.join(self.phones_dir, "disambiguation_symbols.int")
        if os.path.exists(disambig) and os.path.exists(disambig_int):
            return
        with open(disambig, "w", encoding="utf8") as outf, open(
            disambig_int, "w", encoding="utf8"
        ) as intf:
            for d in sorted(self.disambiguation_symbols, key=lambda x: self.phone_mapping[x]):
                outf.write(f"{d}\n")
                intf.write(f"{self.phone_mapping[d]}\n")

    def _write_fst_binary(self, write_disambiguation: Optional[bool] = False) -> None:
        """
        Write the binary fst file to the temporary directory

        Parameters
        ----------
        write_disambiguation: bool, optional
            Flag for including disambiguation symbols
        """
        if write_disambiguation:
            lexicon_fst_path = os.path.join(self.output_directory, "lexicon_disambig.text.fst")
            output_fst = os.path.join(self.output_directory, "L_disambig.fst")
        else:
            lexicon_fst_path = os.path.join(self.output_directory, "lexicon.text.fst")
            output_fst = os.path.join(self.output_directory, "L.fst")
        if os.path.exists(output_fst):
            return

        phones_file_path = os.path.join(self.output_directory, "phones.txt")
        words_file_path = os.path.join(self.output_directory, "words.txt")

        log_path = os.path.join(self.output_directory, "fst.log")
        temp_fst_path = os.path.join(self.output_directory, "temp.fst")
        with open(log_path, "w") as log_file:
            compile_proc = subprocess.Popen(
                [
                    thirdparty_binary("fstcompile"),
                    f"--isymbols={phones_file_path}",
                    f"--osymbols={words_file_path}",
                    "--keep_isymbols=false",
                    "--keep_osymbols=false",
                    lexicon_fst_path,
                    temp_fst_path,
                ],
                stderr=log_file,
            )
            compile_proc.communicate()
            if write_disambiguation:
                temp2_fst_path = os.path.join(self.output_directory, "temp2.fst")
                phone_disambig_path = os.path.join(self.output_directory, "phone_disambig.txt")
                word_disambig_path = os.path.join(self.output_directory, "word_disambig.txt")
                with open(phone_disambig_path, "w") as f:
                    f.write(str(self.phone_mapping["#0"]))
                with open(word_disambig_path, "w") as f:
                    f.write(str(self.words_mapping["#0"]))
                selfloop_proc = subprocess.Popen(
                    [
                        thirdparty_binary("fstaddselfloops"),
                        phone_disambig_path,
                        word_disambig_path,
                        temp_fst_path,
                        temp2_fst_path,
                    ],
                    stderr=log_file,
                )
                selfloop_proc.communicate()
                arc_sort_proc = subprocess.Popen(
                    [
                        thirdparty_binary("fstarcsort"),
                        "--sort_type=olabel",
                        temp2_fst_path,
                        output_fst,
                    ],
                    stderr=log_file,
                )
            else:
                arc_sort_proc = subprocess.Popen(
                    [
                        thirdparty_binary("fstarcsort"),
                        "--sort_type=olabel",
                        temp_fst_path,
                        output_fst,
                    ],
                    stderr=log_file,
                )
            arc_sort_proc.communicate()

    def _write_fst_text(self, write_disambiguation: Optional[bool] = False) -> None:
        """
        Write the text fst file to the temporary directory

        Parameters
        ----------
        write_disambiguation: bool, optional
            Flag for including disambiguation symbols
        """
        if write_disambiguation:
            lexicon_fst_path = os.path.join(self.output_directory, "lexicon_disambig.text.fst")
            sildisambig = f"#{self.max_disambiguation_symbol + 1}"
        else:
            lexicon_fst_path = os.path.join(self.output_directory, "lexicon.text.fst")
        if os.path.exists(lexicon_fst_path):
            return
        if self.sil_prob != 0:
            silphone = self.optional_silence
            nonoptsil = self.nonoptional_silence

            silcost = -1 * math.log(self.sil_prob)
            nosilcost = -1 * math.log(1.0 - self.sil_prob)
            startstate = 0
            loopstate = 1
            silstate = 2
        else:
            loopstate = 0
            nextstate = 1

        with open(lexicon_fst_path, "w", encoding="utf8") as outf:
            if self.sil_prob != 0:
                outf.write(
                    "\t".join(map(str, [startstate, loopstate, "<eps>", "<eps>", nosilcost]))
                    + "\n"
                )  # no silence

                outf.write(
                    "\t".join(map(str, [startstate, loopstate, nonoptsil, "<eps>", silcost]))
                    + "\n"
                )  # silence
                outf.write(
                    "\t".join(map(str, [silstate, loopstate, silphone, "<eps>"])) + "\n"
                )  # no cost
                nextstate = 3
                if write_disambiguation:
                    disambigstate = 3
                    nextstate = 4
                    outf.write(
                        "\t".join(
                            map(str, [startstate, disambigstate, silphone, "<eps>", silcost])
                        )
                        + "\n"
                    )  # silence.
                    outf.write(
                        "\t".join(map(str, [silstate, disambigstate, silphone, "<eps>", silcost]))
                        + "\n"
                    )  # no cost.
                    outf.write(
                        "\t".join(map(str, [disambigstate, loopstate, sildisambig, "<eps>"]))
                        + "\n"
                    )  # silence disambiguation symbol.

            for w in sorted(self.words.keys()):
                if self.exclude_for_alignment(w):
                    continue
                for pron in sorted(
                    self.words[w],
                    key=lambda x: (x["pronunciation"], x["probability"], x["disambiguation"]),
                ):
                    phones = list(pron["pronunciation"])
                    prob = pron["probability"]
                    disambig_symbol = pron["disambiguation"]
                    if self.position_dependent_phones:
                        if len(phones) == 1:
                            phones[0] += "_S"
                        else:
                            for i in range(len(phones)):
                                if i == 0:
                                    phones[i] += "_B"
                                elif i == len(phones) - 1:
                                    phones[i] += "_E"
                                else:
                                    phones[i] += "_I"
                    if not self.pronunciation_probabilities:
                        pron_cost = 0
                    else:
                        if prob is None:
                            prob = 1.0
                        elif not prob:
                            prob = 0.001  # Dithering to ensure low probability entries
                        pron_cost = -1 * math.log(prob)

                    pron_cost_string = ""
                    if pron_cost != 0:
                        pron_cost_string = f"\t{pron_cost}"

                    s = loopstate
                    word_or_eps = w
                    local_nosilcost = nosilcost + pron_cost
                    local_silcost = silcost + pron_cost
                    while len(phones) > 0:
                        p = phones.pop(0)
                        if len(phones) > 0 or (
                            write_disambiguation and disambig_symbol is not None
                        ):
                            ns = nextstate
                            nextstate += 1
                            outf.write(
                                "\t".join(map(str, [s, ns, p, word_or_eps]))
                                + pron_cost_string
                                + "\n"
                            )
                            word_or_eps = "<eps>"
                            pron_cost_string = ""
                            s = ns
                        elif self.sil_prob == 0:
                            ns = loopstate
                            outf.write(
                                "\t".join(map(str, [s, ns, p, word_or_eps]))
                                + pron_cost_string
                                + "\n"
                            )
                            word_or_eps = "<eps>"
                            pron_cost_string = ""
                            s = ns
                        else:
                            outf.write(
                                "\t".join(
                                    map(str, [s, loopstate, p, word_or_eps, local_nosilcost])
                                )
                                + "\n"
                            )
                            outf.write(
                                "\t".join(map(str, [s, silstate, p, word_or_eps, local_silcost]))
                                + "\n"
                            )
                    if write_disambiguation and disambig_symbol is not None:
                        outf.write(
                            "\t".join(
                                map(
                                    str,
                                    [
                                        s,
                                        loopstate,
                                        f"#{disambig_symbol}",
                                        word_or_eps,
                                        local_nosilcost,
                                    ],
                                )
                            )
                            + "\n"
                        )
                        outf.write(
                            "\t".join(
                                map(
                                    str,
                                    [
                                        s,
                                        silstate,
                                        f"#{disambig_symbol}",
                                        word_or_eps,
                                        local_silcost,
                                    ],
                                )
                            )
                            + "\n"
                        )

            outf.write(f"{loopstate}\t0\n")


class MultispeakerDictionary(Dictionary):
    """
    Class containing information about a pronunciation dictionary with different dictionaries per speaker

    Parameters
    ----------
    input_path : str
        Path to an input pronunciation dictionary
    output_directory : str
        Path to a directory to store files for Kaldi
    oov_code : str, optional
        What to label words not in the dictionary, defaults to ``'<unk>'``
    position_dependent_phones : bool, optional
        Specifies whether phones should be represented as dependent on their
        position in the word (beginning, middle or end), defaults to True
    num_sil_states : int, optional
        Number of states to use for silence phones, defaults to 5
    num_nonsil_states : int, optional
        Number of states to use for non-silence phones, defaults to 3
    shared_silence_phones : bool, optional
        Specify whether to share states across all silence phones, defaults
        to True
    sil_prob : float, optional
        Probability of optional silences following words, defaults to 0.5
    word_set : Collection[str], optional
        Word set to limit output files
    debug: bool, optional
        Flag for whether to perform debug steps and prevent intermediate cleanup
    logger: :class:`~logging.Logger`, optional
        Logger to output information to
    punctuation: str, optional
        Punctuation to use when parsing text
    clitic_markers: str, optional
        Clitic markers to use when parsing text
    compound_markers: str, optional
        Compound markers to use when parsing text
    multilingual_ipa: bool, optional
        Flag for multilingual IPA mode, defaults to False
    strip_diacritics: List[str], optional
        Diacritics to strip in multilingual IPA mode
    digraphs: List[str], optional
        Digraphs to split up in multilingual IPA mode
    """

    has_multiple = True

    def __init__(
        self,
        input_path: str,
        output_directory: str,
        oov_code: Optional[str] = "<unk>",
        position_dependent_phones: Optional[bool] = True,
        num_sil_states: Optional[int] = 5,
        num_nonsil_states: Optional[int] = 3,
        shared_silence_phones: Optional[bool] = True,
        sil_prob: Optional[float] = 0.5,
        word_set: Optional[List[str]] = None,
        debug: Optional[bool] = False,
        logger: Optional[Logger] = None,
        punctuation: PunctuationType = None,
        clitic_markers: PunctuationType = None,
        compound_markers: PunctuationType = None,
        multilingual_ipa: Optional[bool] = False,
        strip_diacritics: IpaType = None,
        digraphs: IpaType = None,
    ):
        self.multilingual_ipa = multilingual_ipa
        self.strip_diacritics = DEFAULT_STRIP_DIACRITICS
        self.digraphs = DEFAULT_DIGRAPHS
        if strip_diacritics is not None:
            self.strip_diacritics = strip_diacritics
        if digraphs is not None:
            self.digraphs = digraphs
        self.punctuation = DEFAULT_PUNCTUATION
        self.clitic_markers = DEFAULT_CLITIC_MARKERS
        self.compound_markers = DEFAULT_COMPOUND_MARKERS
        if punctuation is not None:
            self.punctuation = punctuation
        if clitic_markers is not None:
            self.clitic_markers = clitic_markers
        if compound_markers is not None:
            self.compound_markers = compound_markers
        self.input_path = input_path
        self.debug = debug
        self.output_directory = os.path.join(output_directory, "dictionary")
        os.makedirs(self.output_directory, exist_ok=True)
        self.log_file = os.path.join(self.output_directory, "dictionary.log")
        if logger is None:
            self.logger = logging.getLogger("dictionary_setup")
            self.logger.setLevel(logging.INFO)
            handler = logging.FileHandler(self.log_file, "w", "utf-8")
            handler.setFormatter = logging.Formatter("%(name)s %(message)s")
            self.logger.addHandler(handler)
        else:
            self.logger = logger
        self.num_sil_states = num_sil_states
        self.num_nonsil_states = num_nonsil_states
        self.shared_silence_phones = shared_silence_phones
        self.sil_prob = sil_prob
        self.oov_code = oov_code
        self.sil_code = "!sil"
        self.oovs_found = Counter()
        self.position_dependent_phones = position_dependent_phones
        self.max_disambiguation_symbol = 0
        self.disambiguation_symbols = set()
        self.optional_silence = "sp"
        self.nonoptional_silence = "sil"

        if word_set is not None:
            word_set = {sanitize(x, self.punctuation, self.clitic_markers) for x in word_set}
            word_set.add("!sil")
            word_set.add(self.oov_code)
        self.word_set = word_set

        if not os.path.exists(input_path):
            raise (DictionaryPathError(input_path))
        if not os.path.isfile(input_path):
            raise (DictionaryFileError(input_path))

        self.speaker_mapping = {}
        self.dictionary_mapping = {}
        self.logger.info("Parsing multispeaker dictionary file")
        available_langs = get_available_dictionaries()
        with open(input_path, "r", encoding="utf8") as f:
            data = yaml.safe_load(f)
            for speaker, path in data.items():
                if path in available_langs:
                    path = get_dictionary_path(path)
                dictionary_name = os.path.splitext(os.path.basename(path))[0]
                self.speaker_mapping[speaker] = dictionary_name
                if dictionary_name not in self.dictionary_mapping:
                    self.dictionary_mapping[dictionary_name] = Dictionary(
                        path,
                        output_directory,
                        oov_code=self.oov_code,
                        position_dependent_phones=self.position_dependent_phones,
                        word_set=self.word_set,
                        num_sil_states=self.num_sil_states,
                        num_nonsil_states=self.num_nonsil_states,
                        shared_silence_phones=self.shared_silence_phones,
                        sil_prob=self.sil_prob,
                        debug=self.debug,
                        logger=self.logger,
                        punctuation=self.punctuation,
                        clitic_markers=self.clitic_markers,
                        compound_markers=self.compound_markers,
                        multilingual_ipa=self.multilingual_ipa,
                        strip_diacritics=self.strip_diacritics,
                        digraphs=self.digraphs,
                    )

        self.nonsil_phones = set()
        self.sil_phones = {"sp", "spn", "sil"}
        self.words = set()
        self.clitic_set = set()
        for d in self.dictionary_mapping.values():
            self.nonsil_phones.update(d.nonsil_phones)
            self.sil_phones.update(d.sil_phones)
            self.words.update(d.words)
            self.clitic_set.update(d.clitic_set)
        self.words_mapping = {}
        self.phone_mapping = {}

    @property
    def silences(self) -> set:
        """
        Set of silence phones
        """
        return {self.optional_silence, self.nonoptional_silence}

    def get_dictionary_name(self, speaker: Union[str, Speaker]) -> str:
        """
        Get the dictionary name for a given speaker

        Parameters
        ----------
        speaker: Union[Speaker, str]
            Speaker to look up

        Returns
        -------
        str
            Dictionary name for the speaker
        """
        if not isinstance(speaker, str):
            speaker = speaker.name
        if speaker not in self.speaker_mapping:
            return self.speaker_mapping["default"]
        return self.speaker_mapping[speaker]

    def get_dictionary(self, speaker: Union[Speaker, str]) -> Dictionary:
        """
        Get a dictionary for a given speaker

        Parameters
        ----------
        speaker: Union[Speaker, str]
            Speaker to look up

        Returns
        -------
        Dictionary
            Dictionary for the speaker
        """
        return self.dictionary_mapping[self.get_dictionary_name(speaker)]

    def generate_mappings(self) -> None:
        """
        Generate phone and word mappings from text to integer IDs
        """
        self.phone_mapping = {}
        i = 0
        self.phone_mapping["<eps>"] = i
        if self.position_dependent_phones:
            for p in self.positional_sil_phones:
                i += 1
                self.phone_mapping[p] = i
            for p in self.positional_nonsil_phones:
                i += 1
                self.phone_mapping[p] = i
        else:
            for p in sorted(self.sil_phones):
                i += 1
                self.phone_mapping[p] = i
            for p in sorted(self.nonsil_phones):
                i += 1
                self.phone_mapping[p] = i

        self.words_mapping = {}
        i = 0
        self.words_mapping["<eps>"] = i
        for w in sorted(self.words):
            if self.exclude_for_alignment(w):
                continue
            i += 1
            self.words_mapping[w] = i

        self.words_mapping["#0"] = i + 1
        self.words_mapping["<s>"] = i + 2
        self.words_mapping["</s>"] = i + 3
        self.words.update(["<eps>", "#0", "<s>", "</s>"])
        self.oovs_found = Counter()
        self.max_disambiguation_symbol = 0
        for d in self.dictionary_mapping.values():
            d.generate_mappings()
            if d.max_disambiguation_symbol > self.max_disambiguation_symbol:
                self.max_disambiguation_symbol = d.max_disambiguation_symbol
        i = max(self.phone_mapping.values())
        self.disambiguation_symbols = set()
        for x in range(self.max_disambiguation_symbol + 2):
            p = f"#{x}"
            self.disambiguation_symbols.add(p)
            i += 1
            self.phone_mapping[p] = i

    def write(self, write_disambiguation: Optional[bool] = False) -> None:
        """
        Write all child dictionaries to the temporary directory

        Parameters
        ----------
        write_disambiguation: bool, optional
            Flag to use disambiguation symbols in the output
        """
        os.makedirs(self.phones_dir, exist_ok=True)
        self.generate_mappings()
        for d in self.dictionary_mapping.values():
            d.phone_mapping = self.phone_mapping
            d.write(write_disambiguation)

    @property
    def output_paths(self) -> Dict[str, str]:
        """
        Mapping of output directory for child dictionaries
        """
        return {d.name: d.output_directory for d in self.dictionary_mapping.values()}


if TYPE_CHECKING:
    DictionaryType = Union[MultispeakerDictionary, Dictionary]



"""Utility functions for Montreal Forced Aligner"""
from __future__ import annotations

import logging
import os
import shutil
import sys
import textwrap
from typing import TYPE_CHECKING, Any, Dict, List, Union

import yaml
from colorama import Fore, Style

from .exceptions import KaldiProcessingError, ThirdpartyError
from .models import MODEL_TYPES

if TYPE_CHECKING:
    from .config.base_config import BaseConfig

__all__ = [
    "thirdparty_binary",
    "get_available_dictionaries",
    "log_config",
    "log_kaldi_errors",
    "get_available_models",
    "get_available_language_models",
    "get_available_acoustic_models",
    "get_available_g2p_models",
    "get_pretrained_language_model_path",
    "get_pretrained_g2p_path",
    "get_pretrained_ivector_path",
    "get_pretrained_path",
    "get_pretrained_acoustic_path",
    "get_dictionary_path",
    "get_available_ivector_extractors",
    "guess_model_type",
    "parse_logs",
    "setup_logger",
    "CustomFormatter",
]


def get_mfa_version():
    try:
        from .version import version as __version__  # noqa
    except ImportError:
        __version__ = "2.0.0"
    return __version__


def thirdparty_binary(binary_name: str) -> str:
    """
    Generate full path to a given binary name

    Notes
    -----
    With the move to conda, this function is deprecated as conda will manage the path much better

    Parameters
    ----------
    binary_name: str
        Executable to run

    Returns
    -------
    str
        Full path to the executable
    """
    bin_path = shutil.which(binary_name)
    if bin_path is None:
        if binary_name in ["fstcompile", "fstarcsort", "fstconvert"] and sys.platform != "win32":
            raise ThirdpartyError(binary_name, open_fst=True)
        else:
            raise ThirdpartyError(binary_name)
    return bin_path


def parse_logs(log_directory: str) -> None:
    """
    Parse the output of a Kaldi run for any errors and raise relevant MFA exceptions

    Parameters
    ----------
    log_directory: str
        Log directory to parse

    Raises
    ------
    KaldiProcessingError
        If any log files contained error lines

    """
    error_logs = []
    for name in os.listdir(log_directory):
        log_path = os.path.join(log_directory, name)
        with open(log_path, "r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if "error while loading shared libraries: libopenblas.so.0" in line:
                    raise ThirdpartyError("libopenblas.so.0", open_blas=True)
                for libc_version in ["GLIBC_2.27", "GLIBCXX_3.4.20"]:
                    if libc_version in line:
                        raise ThirdpartyError(libc_version, libc=True)
                if "sox FAIL formats" in line:
                    f = line.split(" ")[-1]
                    raise ThirdpartyError(f, sox=True)
                if line.startswith("ERROR") or line.startswith("ASSERTION_FAILED"):
                    error_logs.append(log_path)
                    break
    if error_logs:
        raise KaldiProcessingError(error_logs)


def log_kaldi_errors(error_logs: List[str], logger: logging.Logger) -> None:
    """
    Save details of Kaldi processing errors to a logger

    Parameters
    ----------
    error_logs: List[str]
        Kaldi log files with errors
    logger: :class:`~logging.Logger`
        Logger to output to
    """
    logger.debug("There were {} kaldi processing files that had errors:".format(len(error_logs)))
    for path in error_logs:
        logger.debug("")
        logger.debug(path)
        with open(path, "r", encoding="utf8") as f:
            for line in f:
                logger.debug("\t" + line.strip())


def get_available_models(model_type: str) -> List[str]:
    """
    Get a list of available models for a given model type

    Parameters
    ----------
    model_type: str
        Model type to search

    Returns
    -------
    List[str]
        List of model names
    """
    from .config import TEMP_DIR

    pretrained_dir = os.path.join(TEMP_DIR, "pretrained_models", model_type)
    os.makedirs(pretrained_dir, exist_ok=True)
    available = []
    model_class = MODEL_TYPES[model_type]
    for f in os.listdir(pretrained_dir):
        if model_class is None or model_class.valid_extension(f):
            available.append(os.path.splitext(f)[0])
    return available


def guess_model_type(path: str) -> List[str]:
    """
    Guess a model type given a path

    Parameters
    ----------
    path: str
        Model archive to guess

    Returns
    -------
    List[str]
        Possible model types that use that extension
    """
    ext = os.path.splitext(path)[1]
    if not ext:
        return []
    possible = []
    for m, mc in MODEL_TYPES.items():
        if ext in mc.extensions:
            possible.append(m)
    return possible


def get_available_acoustic_models() -> List[str]:
    """
    Return a list of all available acoustic models

    Returns
    -------
    List[str]
        Pretrained acoustic models
    """
    return get_available_models("acoustic")


def get_available_g2p_models() -> List[str]:
    """
    Return a list of all available G2P models

    Returns
    -------
    List[str]
        Pretrained G2P models
    """
    return get_available_models("g2p")


def get_available_ivector_extractors() -> List[str]:
    """
    Return a list of all available ivector extractors

    Returns
    -------
    List[str]
        Pretrained ivector extractors
    """
    return get_available_models("ivector")


def get_available_language_models() -> List[str]:
    """
    Return a list of all available language models

    Returns
    -------
    List[str]
        Pretrained language models
    """
    return get_available_models("language_model")


def get_available_dictionaries() -> List[str]:
    """
    Return a list of all available dictionaries

    Returns
    -------
    List[str]
        Saved dictionaries
    """
    return get_available_models("dictionary")


def get_pretrained_path(model_type: str, name: str, enforce_existence: bool = True) -> str:
    """
    Generate a path to a pretrained model based on its name and model type

    Parameters
    ----------
    model_type: str
        Type of model
    name: str
        Name of model
    enforce_existence: bool
        Flag to return None if the path doesn't exist, defaults to True

    Returns
    -------
    str
        Path to model
    """
    from .config import TEMP_DIR

    pretrained_dir = os.path.join(TEMP_DIR, "pretrained_models", model_type)
    model_class = MODEL_TYPES[model_type]
    return model_class.generate_path(pretrained_dir, name, enforce_existence)


def get_pretrained_acoustic_path(name: str) -> str:
    """
    Generate a path to a given pretrained acoustic model

    Parameters
    ----------
    name: str
        Name of model

    Returns
    -------
    str
        Full path to model
    """
    return get_pretrained_path("acoustic", name)


def get_pretrained_ivector_path(name: str) -> str:
    """
    Generate a path to a given pretrained ivector extractor

    Parameters
    ----------
    name: str
        Name of model

    Returns
    -------
    str
        Full path to model
    """
    return get_pretrained_path("ivector", name)


def get_pretrained_language_model_path(name: str) -> str:
    """
    Generate a path to a given pretrained language model

    Parameters
    ----------
    name: str
        Name of model

    Returns
    -------
    str
        Full path to model
    """
    return get_pretrained_path("language_model", name)


def get_pretrained_g2p_path(name: str) -> str:
    """
    Generate a path to a given pretrained G2P model

    Parameters
    ----------
    name: str
        Name of model

    Returns
    -------
    str
        Full path to model
    """
    return get_pretrained_path("g2p", name)


def get_dictionary_path(name: str) -> str:
    """
    Generate a path to a given saved dictionary

    Parameters
    ----------
    name: str
        Name of dictionary

    Returns
    -------
    str
        Full path to dictionary
    """
    return get_pretrained_path("dictionary", name)


class CustomFormatter(logging.Formatter):
    """
    Custom log formatter class for MFA to highlight messages and incorporate terminal options from
    the global configuration
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from .config import load_global_config

        config = load_global_config()
        self.width = config["terminal_width"]
        use_colors = config.get("terminal_colors", True)
        red = ""
        green = ""
        yellow = ""
        blue = ""
        reset = ""
        if use_colors:
            red = Fore.RED
            green = Fore.GREEN
            yellow = Fore.YELLOW
            blue = Fore.CYAN
            reset = Style.RESET_ALL

        self.FORMATS = {
            logging.DEBUG: (f"{blue}DEBUG{reset} - ", "%(message)s"),
            logging.INFO: (f"{green}INFO{reset} - ", "%(message)s"),
            logging.WARNING: (f"{yellow}WARNING{reset} - ", "%(message)s"),
            logging.ERROR: (f"{red}ERROR{reset} - ", "%(message)s"),
            logging.CRITICAL: (f"{red}CRITICAL{reset} - ", "%(message)s"),
        }

    def format(self, record: logging.LogRecord):
        """
        Format a given log message

        Parameters
        ----------
        record: logging.LogRecord
            Log record to format

        Returns
        -------
        str
            Formatted log message
        """
        log_fmt = self.FORMATS.get(record.levelno)
        return textwrap.fill(
            record.getMessage(),
            initial_indent=log_fmt[0],
            subsequent_indent=" " * len(log_fmt[0]),
            width=self.width,
        )


def setup_logger(
    identifier: str, output_directory: str, console_level: str = "info"
) -> logging.Logger:
    """
    Construct a logger for a command line run

    Parameters
    ----------
    identifier: str
        Name of the MFA utility
    output_directory: str
        Top level logging directory
    console_level: str, optional
        Level to output to the console, defaults to "info"

    Returns
    -------
    :class:`~logging.Logger`
        Logger to use
    """
    os.makedirs(output_directory, exist_ok=True)
    log_path = os.path.join(output_directory, f"{identifier}.log")
    if os.path.exists(log_path):
        os.remove(log_path)
    logger = logging.getLogger(identifier)
    logger.setLevel(logging.DEBUG)

    handler = logging.FileHandler(log_path, encoding="utf8")
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(getattr(logging, console_level.upper()))
    handler.setFormatter(CustomFormatter())
    logger.addHandler(handler)

    return logger


def log_config(logger: logging.Logger, config: Union[Dict[str, Any], BaseConfig]) -> None:
    """
    Output a configuration to a Logger

    Parameters
    ----------
    logger: :class:`~logging.Logger`
        Logger to save to
    config: Dict[str, Any]
        Configuration to dump
    """
    stream = yaml.dump(config)
    logger.debug(stream)



"""Class definitions for configuring G2P model training"""
from __future__ import annotations

import yaml

from .base_config import (
    DEFAULT_CLITIC_MARKERS,
    DEFAULT_COMPOUND_MARKERS,
    DEFAULT_PUNCTUATION,
    BaseConfig,
    ConfigError,
)

__all__ = ["TrainG2PConfig", "train_g2p_yaml_to_config", "load_basic_train_g2p_config"]


class TrainG2PConfig(BaseConfig):
    """
    Configuration class for training G2P models
    """

    def __init__(self):
        self.punctuation = DEFAULT_PUNCTUATION
        self.clitic_markers = DEFAULT_CLITIC_MARKERS
        self.compound_markers = DEFAULT_COMPOUND_MARKERS
        self.num_pronunciations = 1
        self.order = 7
        self.random_starts = 25
        self.seed = 1917
        self.delta = 1 / 1024
        self.lr = 1.0
        self.batch_size = 200
        self.max_iterations = 10
        self.smoothing_method = "kneser_ney"
        self.pruning_method = "relative_entropy"
        self.model_size = 1000000
        self.use_mp = True

    def update(self, data: dict) -> None:
        """Update configuration parameters"""
        for k, v in data.items():
            if k in ["punctuation", "clitic_markers", "compound_markers"]:
                if not v:
                    continue
                if "-" in v:
                    v = "-" + v.replace("-", "")
                if "]" in v and r"\]" not in v:
                    v = v.replace("]", r"\]")
            elif not hasattr(self, k):
                raise ConfigError("No field found for key {}".format(k))
            setattr(self, k, v)


def train_g2p_yaml_to_config(path: str) -> TrainG2PConfig:
    """
    Helper function to load G2P training configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_g2p_config.TrainG2PConfig`
        G2P training configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        global_params = {}
        for k, v in data.items():
            global_params[k] = v
        g2p_config = TrainG2PConfig()
        g2p_config.update(global_params)
        return g2p_config


def load_basic_train_g2p_config() -> TrainG2PConfig:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_g2p_config.TrainG2PConfig`
        Default G2P training configuration
    """
    return TrainG2PConfig()



"""Class definitions for configuring MFA"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List

if TYPE_CHECKING:
    from argparse import Namespace

    ConfigDict = Dict[str, Any]
import os

import yaml

from .align_config import AlignConfig, align_yaml_to_config, load_basic_align  # noqa
from .command_config import load_command_configuration  # noqa
from .feature_config import FeatureConfig  # noqa
from .segmentation_config import (  # noqa
    SegmentationConfig,
    load_basic_segmentation,
    segmentation_yaml_to_config,
)
from .speaker_classification_config import (  # noqa
    SpeakerClassificationConfig,
    classification_yaml_to_config,
    load_basic_classification,
)
from .train_config import (  # noqa
    TrainingConfig,
    load_basic_train,
    load_basic_train_ivector,
    load_test_config,
    train_yaml_to_config,
)
from .train_lm_config import TrainLMConfig, load_basic_train_lm, train_lm_yaml_to_config  # noqa
from .transcribe_config import (  # noqa
    TranscribeConfig,
    load_basic_transcribe,
    transcribe_yaml_to_config,
)

TEMP_DIR = os.path.expanduser("~/Documents/MFA")


def generate_config_path() -> str:
    """
    Generate the global configuration path for MFA

    Returns
    -------
    str
        Full path to configuration yaml
    """
    return os.path.join(TEMP_DIR, "global_config.yaml")


def generate_command_history_path() -> str:
    """
    Generate the path to the command history file

    Returns
    -------
    str
        Full path to history file
    """
    return os.path.join(TEMP_DIR, "command_history.yaml")


def load_command_history() -> List[str]:
    """
    Load command history for MFA

    Returns
    -------
    List
        List of commands previously run
    """
    path = generate_command_history_path()
    history = []
    if os.path.exists(path):
        with open(path, "r", encoding="utf8") as f:
            history = yaml.safe_load(f)
    return history


def update_command_history(command_data: dict) -> None:
    """
    Update command history with most recent command

    Parameters
    ----------
    command_data: dict
        Current command metadata
    """
    try:
        if command_data["command"].split(" ")[1] == "history":
            return
    except Exception:
        return
    history = load_command_history()
    path = generate_command_history_path()
    history.append(command_data)
    history = history[-50:]
    with open(path, "w", encoding="utf8") as f:
        yaml.safe_dump(history, f)


def update_global_config(args: Namespace) -> None:
    """
    Update global configuration of MFA

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Arguments to set
    """
    global_configuration_file = generate_config_path()
    default_config = {
        "clean": False,
        "verbose": False,
        "debug": False,
        "overwrite": False,
        "terminal_colors": True,
        "terminal_width": 120,
        "cleanup_textgrids": True,
        "num_jobs": 3,
        "blas_num_threads": 1,
        "use_mp": True,
        "temp_directory": TEMP_DIR,
    }
    if os.path.exists(global_configuration_file):
        with open(global_configuration_file, "r", encoding="utf8") as f:
            data = yaml.safe_load(f)
            default_config.update(data)
    if args.always_clean:
        default_config["clean"] = True
    if args.never_clean:
        default_config["clean"] = False
    if args.always_verbose:
        default_config["verbose"] = True
    if args.never_verbose:
        default_config["verbose"] = False
    if args.always_debug:
        default_config["debug"] = True
    if args.never_debug:
        default_config["debug"] = False
    if args.always_overwrite:
        default_config["overwrite"] = True
    if args.never_overwrite:
        default_config["overwrite"] = False
    if args.disable_mp:
        default_config["use_mp"] = False
    if args.enable_mp:
        default_config["use_mp"] = True
    if args.disable_textgrid_cleanup:
        default_config["cleanup_textgrids"] = False
    if args.enable_textgrid_cleanup:
        default_config["cleanup_textgrids"] = True
    if args.disable_terminal_colors:
        default_config["terminal_colors"] = False
    if args.enable_terminal_colors:
        default_config["terminal_colors"] = True
    if args.num_jobs and args.num_jobs > 0:
        default_config["num_jobs"] = args.num_jobs
    if args.terminal_width and args.terminal_width > 0:
        default_config["terminal_width"] = args.terminal_width
    if args.blas_num_threads and args.blas_num_threads > 0:
        default_config["blas_num_threads"] = args.blas_num_threads
    if args.temp_directory:
        default_config["temp_directory"] = args.temp_directory
    with open(global_configuration_file, "w", encoding="utf8") as f:
        yaml.dump(default_config, f)


def load_global_config() -> Dict[str, Any]:
    """
    Load the global MFA configuration

    Returns
    -------
    Dict
        Global configuration
    """
    global_configuration_file = generate_config_path()
    default_config = {
        "clean": False,
        "verbose": False,
        "debug": False,
        "overwrite": False,
        "terminal_colors": True,
        "terminal_width": 120,
        "cleanup_textgrids": True,
        "num_jobs": 3,
        "blas_num_threads": 1,
        "use_mp": True,
        "temp_directory": TEMP_DIR,
    }
    if os.path.exists(global_configuration_file):
        with open(global_configuration_file, "r", encoding="utf8") as f:
            data = yaml.safe_load(f)
            default_config.update(data)
    return default_config


USE_COLORS = load_global_config().get("terminal_colors", True)
BLAS_THREADS = load_global_config().get("blas_num_threads", 1)



"""Classes for training language models"""
from __future__ import annotations

import logging
import os
import re
import subprocess
from typing import TYPE_CHECKING, Dict, Optional, Union

from ..config import TEMP_DIR
from ..corpus import Corpus
from ..models import LanguageModel

if TYPE_CHECKING:
    from ..config.train_lm_config import TrainLMConfig
    from ..dictionary import DictionaryType


__all__ = ["LmTrainer"]


class LmTrainer:
    """
    Train a language model from a corpus with text, or convert an existing ARPA-format language model to MFA format

    Parameters
    ----------
    source: class:`~montreal_forced_aligner.corpus.base.Corpus` or str
        Either a alignable corpus or a path to an ARPA format language model
    config : class:`~montreal_forced_aligner.config.TrainLMConfig`
        Config class for training language model
    output_model_path : str
        Path to output trained model
    dictionary : class:`~montreal_forced_aligner.dictionary.Dictionary`, optional
        Optional dictionary to calculate unknown words
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    supplemental_model_path: str, optional
        Path to second language model to merge with the trained model
    supplemental_model_weight : float, optional
        Weight of supplemental model when merging, defaults to 1
    debug : bool
        Flag for debug mode
    logger : :class:`~logging.Logger`, optional
        Logger to send output to
    """

    def __init__(
        self,
        source: Union[Corpus, str],
        config: TrainLMConfig,
        output_model_path: str,
        dictionary: Optional[DictionaryType] = None,
        temp_directory: Optional[str] = None,
        supplemental_model_path: Optional[str] = None,
        supplemental_model_weight: int = 1,
        debug: bool = False,
        logger: Optional[logging.Logger] = None,
    ):
        if not temp_directory:
            temp_directory = TEMP_DIR
        temp_directory = os.path.join(temp_directory, "LM")
        self.debug = debug
        self.name, _ = os.path.splitext(os.path.basename(output_model_path))
        self.temp_directory = os.path.join(temp_directory, self.name)
        self.models_temp_dir = os.path.join(self.temp_directory, "models")
        self.log_directory = os.path.join(self.temp_directory, "logs")
        self.log_file = os.path.join(self.log_directory, "train_lm.log")
        os.makedirs(self.log_directory, exist_ok=True)
        if logger is None:
            self.logger = logging.getLogger("train_lm")
            self.logger.setLevel(logging.INFO)
            handler = logging.FileHandler(self.log_file, "w", "utf-8")
            handler.setFormatter = logging.Formatter("%(name)s %(message)s")
            self.logger.addHandler(handler)
        else:
            self.logger = logger
        self.source = source
        self.dictionary = dictionary
        self.output_model_path = output_model_path
        self.config = config
        self.supplemental_model_path = supplemental_model_path
        self.source_model_weight = 1
        self.supplemental_model_weight = supplemental_model_weight

    @property
    def meta(self) -> Dict[str, Union[str, int, float]]:
        """Metadata information for the language model"""
        from ..utils import get_mfa_version

        return {
            "type": "ngram",
            "order": self.config.order,
            "method": self.config.method,
            "prune": self.config.prune,
            "version": get_mfa_version(),
        }

    def evaluate(self) -> None:
        """
        Run an evaluation over the training data to generate perplexity score
        """
        log_path = os.path.join(self.log_directory, "evaluate.log")
        mod_path = os.path.join(self.temp_directory, self.name + ".mod")
        far_path = os.path.join(self.temp_directory, self.name + ".far")
        small_mod_path = mod_path.replace(".mod", "_small.mod")
        med_mod_path = mod_path.replace(".mod", "_med.mod")
        with open(log_path, "w", encoding="utf8") as log_file:
            perplexity_proc = subprocess.Popen(
                ["ngramperplexity", '--OOV_symbol="<unk>"', mod_path, far_path],
                stdout=subprocess.PIPE,
                stderr=log_file,
                text=True,
            )
            stdout, stderr = perplexity_proc.communicate()
            num_sentences = None
            num_words = None
            num_oovs = None
            perplexity = None
            for line in stdout.splitlines():
                m = re.search(r"(\d+) sentences", line)
                if m:
                    num_sentences = m.group(0)
                m = re.search(r"(\d+) words", line)
                if m:
                    num_words = m.group(0)
                m = re.search(r"(\d+) OOVs", line)
                if m:
                    num_oovs = m.group(0)
                m = re.search(r"perplexity = ([\d.]+)", line)
                if m:
                    perplexity = m.group(0)

            self.logger.info(f"{num_sentences} sentences, {num_words} words, {num_oovs} oovs")
            self.logger.info(f"Perplexity of large model: {perplexity}")

            perplexity_proc = subprocess.Popen(
                ["ngramperplexity", '--OOV_symbol="<unk>"', med_mod_path, far_path],
                stdout=subprocess.PIPE,
                stderr=log_file,
                text=True,
            )
            stdout, stderr = perplexity_proc.communicate()

            perplexity = None
            for line in stdout.splitlines():
                m = re.search(r"perplexity = ([\d.]+)", line)
                if m:
                    perplexity = m.group(0)
            self.logger.info(f"Perplexity of medium model: {perplexity}")
            perplexity_proc = subprocess.Popen(
                ["ngramperplexity", '--OOV_symbol="<unk>"', small_mod_path, far_path],
                stdout=subprocess.PIPE,
                stderr=log_file,
                text=True,
            )
            stdout, stderr = perplexity_proc.communicate()

            perplexity = None
            for line in stdout.splitlines():
                m = re.search(r"perplexity = ([\d.]+)", line)
                if m:
                    perplexity = m.group(0)
            self.logger.info(f"Perplexity of small model: {perplexity}")

    def train(self) -> None:
        """
        Train a language model
        """
        mod_path = os.path.join(self.temp_directory, f"{self.name}.mod")
        large_model_path = os.path.join(self.temp_directory, f"{self.name}.arpa")
        small_output_path = large_model_path.replace(".arpa", "_small.arpa")
        med_output_path = large_model_path.replace(".arpa", "_med.arpa")
        if isinstance(self.source, Corpus):
            self.logger.info("Beginning training large ngram model...")
            sym_path = os.path.join(self.temp_directory, f"{self.name}.sym")
            far_path = os.path.join(self.temp_directory, f"{self.name}.far")
            cnts_path = os.path.join(self.temp_directory, f"{self.name}.cnts")
            training_path = os.path.join(self.temp_directory, "training.txt")

            with open(training_path, "w", encoding="utf8") as f:
                for text in self.source.normalized_text_iter(
                    self.dictionary, self.config.count_threshold
                ):
                    f.write(f"{text}\n")

            if self.dictionary is not None:
                self.dictionary.save_oovs_found(self.temp_directory)

            subprocess.call(["ngramsymbols", '--OOV_symbol="<unk>"', training_path, sym_path])
            subprocess.call(
                [
                    "farcompilestrings",
                    "--fst_type=compact",
                    '--unknown_symbol="<unk>"',
                    "--symbols=" + sym_path,
                    "--keep_symbols",
                    training_path,
                    far_path,
                ]
            )
            subprocess.call(["ngramcount", f"--order={self.config.order}", far_path, cnts_path])
            subprocess.call(["ngrammake", f"--method={self.config.method}", cnts_path, mod_path])
            self.logger.info("Done!")
        else:
            self.logger.info("Parsing large ngram model...")
            temp_text_path = os.path.join(self.temp_directory, "input.arpa")
            with open(self.source, "r", encoding="utf8") as inf, open(
                temp_text_path, "w", encoding="utf8"
            ) as outf:
                for line in inf:
                    outf.write(line.lower())
            subprocess.call(["ngramread", "--ARPA", temp_text_path, mod_path])
            os.remove(temp_text_path)
        if self.supplemental_model_path:
            self.logger.info("Parsing supplemental ngram model...")
            supplemental_path = os.path.join(self.temp_directory, "extra.mod")
            merged_path = os.path.join(self.temp_directory, "merged.mod")
            subprocess.call(
                ["ngramread", "--ARPA", self.supplemental_model_path, supplemental_path]
            )
            self.logger.info("Merging both ngram models to create final large model...")
            subprocess.call(
                [
                    "ngrammerge",
                    "--normalize",
                    f"--alpha={self.source_model_weight}",
                    f"--beta={self.supplemental_model_weight}",
                    mod_path,
                    supplemental_path,
                    merged_path,
                ]
            )
            mod_path = merged_path

        subprocess.call(["ngramprint", "--ARPA", mod_path, large_model_path])

        self.logger.info("Large ngam model created!")
        directory, filename = os.path.split(self.output_model_path)
        basename, _ = os.path.splitext(filename)

        if self.config.prune:
            self.logger.info("Pruning large ngram model to medium and small versions...")
            small_mod_path = mod_path.replace(".mod", "_small.mod")
            med_mod_path = mod_path.replace(".mod", "_med.mod")
            subprocess.call(
                [
                    "ngramshrink",
                    "--method=relative_entropy",
                    f"--theta={self.config.prune_thresh_small}",
                    mod_path,
                    small_mod_path,
                ]
            )
            subprocess.call(
                [
                    "ngramshrink",
                    "--method=relative_entropy",
                    f"--theta={self.config.prune_thresh_medium}",
                    mod_path,
                    med_mod_path,
                ]
            )
            subprocess.call(["ngramprint", "--ARPA", small_mod_path, small_output_path])
            subprocess.call(["ngramprint", "--ARPA", med_mod_path, med_output_path])
            self.logger.info("Done!")
        self.evaluate()
        model = LanguageModel.empty(basename, root_directory=self.models_temp_dir)
        model.add_meta_file(self)
        model.add_arpa_file(large_model_path)
        if self.config.prune:
            model.add_arpa_file(med_output_path)
            model.add_arpa_file(small_output_path)
        basename, _ = os.path.splitext(self.output_model_path)
        model.dump(basename)
        # model.clean_up()



"""Command line functions for training G2P models"""
from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING, Optional

from montreal_forced_aligner.command_line.utils import validate_model_arg
from montreal_forced_aligner.config import TEMP_DIR
from montreal_forced_aligner.config.train_g2p_config import (
    load_basic_train_g2p_config,
    train_g2p_yaml_to_config,
)
from montreal_forced_aligner.dictionary import Dictionary
from montreal_forced_aligner.g2p.trainer import PyniniTrainer as Trainer

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = ["train_g2p", "validate_args", "run_train_g2p"]


def train_g2p(args: Namespace, unknown_args: Optional[list] = None) -> None:
    """
    Run the G2P model training

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: List[str]
        Optional arguments that will be passed to configuration objects
    """
    if not args.temp_directory:
        temp_dir = TEMP_DIR
    else:
        temp_dir = os.path.expanduser(args.temp_directory)
    if args.clean:
        shutil.rmtree(os.path.join(temp_dir, "G2P"), ignore_errors=True)
        shutil.rmtree(os.path.join(temp_dir, "models", "G2P"), ignore_errors=True)
    if args.config_path:
        train_config = train_g2p_yaml_to_config(args.config_path)
    else:
        train_config = load_basic_train_g2p_config()
    train_config.use_mp = not args.disable_mp
    if unknown_args:
        train_config.update_from_unknown_args(unknown_args)
    dictionary = Dictionary(args.dictionary_path, "")
    t = Trainer(
        dictionary,
        args.output_model_path,
        temp_directory=temp_dir,
        train_config=train_config,
        num_jobs=args.num_jobs,
        verbose=args.verbose,
    )
    if args.validate:
        t.validate()
    else:
        t.train()


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    args.dictionary_path = validate_model_arg(args.dictionary_path, "dictionary")


def run_train_g2p(args: Namespace, unknown: Optional[list] = None) -> None:
    """
    Wrapper function for running G2P model training

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: List[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    train_g2p(args, unknown)



"""Class definitions for configuring language model training"""
from __future__ import annotations

import os

import yaml

from .base_config import BaseConfig

__all__ = ["TrainLMConfig", "train_lm_yaml_to_config", "load_basic_train_lm"]


class TrainLMConfig(BaseConfig):
    """
    Class for storing configuration information for training language models

    Attributes
    ----------
    order: int
    method: str
    prune: bool
    count_threshold: int
    prune_thresh_small: float
    prune_thresh_medium: float
    use_mp: bool
    """

    def __init__(self):
        self.order = 3
        self.method = "kneser_ney"
        self.prune = False
        self.count_threshold = 1
        self.prune_thresh_small = 0.0000003
        self.prune_thresh_medium = 0.0000001
        self.use_mp = True


def train_lm_yaml_to_config(path: str) -> TrainLMConfig:
    """
    Helper function to load language model training configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_lm_config.TrainLMConfig`
        Language model training configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        config = TrainLMConfig()
        config.update(data)
    return config


def load_basic_train_lm() -> TrainLMConfig:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.train_lm_config.TrainLMConfig`
        Default language model training configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    training_config = train_lm_yaml_to_config(os.path.join(base_dir, "basic_train_lm.yaml"))
    return training_config



"""Multiprocessing functions for generating pronunciation probabilities"""
from __future__ import annotations

import os
import subprocess
from collections import Counter, defaultdict
from typing import TYPE_CHECKING, Dict, List, Tuple

from ..utils import thirdparty_binary
from .helper import run_mp, run_non_mp

if TYPE_CHECKING:
    from .alignment import AlignerType


__all__ = ["generate_pronunciations"]


def generate_pronunciations_func(
    log_path: str,
    dictionaries: List[str],
    text_int_paths: Dict[str, str],
    word_boundary_paths: Dict[str, str],
    ali_paths: Dict[str, str],
    model_path: str,
    pron_paths: Dict[str, str],
):
    """
    Multiprocessing function for generating pronunciations

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: List[str]
        List of dictionary names
    text_int_paths: Dict[str, str]
        Dictionary of text int files per dictionary name
    word_boundary_paths: Dict[str, str]
        Dictionary of word boundary files per dictionary name
    ali_paths: Dict[str, str]
        Dictionary of alignment archives per dictionary name
    model_path: str
        Path to acoustic model file
    pron_paths: Dict[str, str]
        Dictionary of pronunciation archives per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            text_int_path = text_int_paths[dict_name]
            word_boundary_path = word_boundary_paths[dict_name]
            ali_path = ali_paths[dict_name]
            pron_path = pron_paths[dict_name]

            lin_proc = subprocess.Popen(
                [
                    thirdparty_binary("linear-to-nbest"),
                    f"ark:{ali_path}",
                    f"ark:{text_int_path}",
                    "",
                    "",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            align_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-align-words"),
                    word_boundary_path,
                    model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=lin_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )

            prons_proc = subprocess.Popen(
                [thirdparty_binary("nbest-to-prons"), model_path, "ark:-", pron_path],
                stdin=align_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            prons_proc.communicate()


def generate_pronunciations(
    aligner: AlignerType,
) -> Tuple[Dict[str, defaultdict[Counter]], Dict[str, Dict[str, List[str, ...]]]]:
    """
    Generates pronunciations based on alignments for a corpus and calculates pronunciation probabilities

    Parameters
    ----------
    aligner: :class:`~montreal_forced_aligner.aligner.pretrained.PretrainedAligner`
        Aligner
    """
    jobs = [x.generate_pronunciations_arguments(aligner) for x in aligner.corpus.jobs]
    if aligner.align_config.use_mp:
        run_mp(generate_pronunciations_func, jobs, aligner.working_log_directory)
    else:
        run_non_mp(generate_pronunciations_func, jobs, aligner.working_log_directory)
    pron_counts = {}
    utt_mapping = {}
    for j in aligner.corpus.jobs:
        args = j.generate_pronunciations_arguments(aligner)
        dict_data = j.dictionary_data()
        for dict_name, pron_path in args.pron_paths.items():
            if dict_name not in pron_counts:
                pron_counts[dict_name] = defaultdict(Counter)
                utt_mapping[dict_name] = {}
            word_lookup = dict_data[dict_name].reversed_words_mapping
            phone_lookup = dict_data[dict_name].reversed_phone_mapping
            with open(pron_path, "r", encoding="utf8") as f:
                last_utt = None
                for line in f:
                    line = line.split()
                    utt = line[0]
                    if utt not in utt_mapping[dict_name]:
                        if last_utt is not None:
                            utt_mapping[dict_name][last_utt].append("</s>")
                        utt_mapping[dict_name][utt] = ["<s>"]
                        last_utt = utt

                    word = word_lookup[int(line[3])]
                    if word == "<eps>":
                        utt_mapping[dict_name][utt].append(word)
                    else:
                        pron = tuple(phone_lookup[int(x)].split("_")[0] for x in line[4:])
                        pron_string = " ".join(pron)
                        utt_mapping[dict_name][utt].append(word + " " + pron_string)
                        pron_counts[dict_name][word][pron] += 1
    return pron_counts, utt_mapping



"""Class definition for MFA's validator"""
from __future__ import annotations

import logging
import os
from decimal import Decimal
from typing import TYPE_CHECKING, Optional

from .aligner.pretrained import PretrainedAligner
from .config import FeatureConfig
from .exceptions import CorpusError, KaldiProcessingError
from .helper import edit_distance, load_scp
from .multiprocessing import run_mp, run_non_mp
from .multiprocessing.alignment import compile_utterance_train_graphs_func, test_utterances_func
from .trainers import MonophoneTrainer
from .utils import log_kaldi_errors

if TYPE_CHECKING:
    from .corpus import Corpus
    from .dictionary import DictionaryType


__all__ = ["CorpusValidator"]


class CorpusValidator:
    """
    Aligner that aligns and trains acoustics models on a large dataset

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus object for the dataset
    dictionary : :class:`~montreal_forced_aligner.dictionary.Dictionary`
        Dictionary object for the pronunciation dictionary
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    ignore_acoustics: bool, optional
        Flag for whether all acoustics should be ignored, which speeds up the validation, defaults to False
    test_transcriptions: bool, optional
        Flag for whether the validator should test transcriptions, defaults to False
    use_mp: bool, optional
        Flag for whether to use multiprocessing
    logger: :class:`~logging.Logger`, optional
        Logger to use
    """

    corpus_analysis_template = """
    =========================================Corpus=========================================
    {} sound files
    {} sound files with .lab transcription files
    {} sound files with TextGrids transcription files
    {} additional sound files ignored (see below)
    {} speakers
    {} utterances
    {} seconds total duration

    DICTIONARY
    ----------
    {}

    SOUND FILE READ ERRORS
    ----------------------
    {}

    FEATURE CALCULATION
    -------------------
    {}

    FILES WITHOUT TRANSCRIPTIONS
    ----------------------------
    {}

    TRANSCRIPTIONS WITHOUT FILES
    --------------------
    {}

    TEXTGRID READ ERRORS
    --------------------
    {}

    UNREADABLE TEXT FILES
    --------------------
    {}
    """

    alignment_analysis_template = """
    =======================================Alignment========================================
    {}
    """

    transcription_analysis_template = """
    ====================================Transcriptions======================================
    {}
    """

    def __init__(
        self,
        corpus: Corpus,
        dictionary: DictionaryType,
        temp_directory: Optional[str] = None,
        ignore_acoustics: bool = False,
        test_transcriptions: bool = False,
        use_mp: bool = True,
        logger: Optional[logging.Logger] = None,
    ):
        self.dictionary = dictionary
        self.corpus = corpus
        self.temp_directory = temp_directory
        self.test_transcriptions = test_transcriptions
        self.ignore_acoustics = ignore_acoustics
        self.trainer: MonophoneTrainer = MonophoneTrainer(FeatureConfig())
        self.logger = logger
        self.trainer.logger = logger
        self.trainer.update({"use_mp": use_mp})
        self.setup()

    def setup(self):
        """
        Set up the corpus and validator

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        self.dictionary.set_word_set(self.corpus.word_set)
        self.dictionary.write()
        if self.test_transcriptions:
            self.dictionary.write(write_disambiguation=True)
        if self.ignore_acoustics:
            fc = None
            if self.logger is not None:
                self.logger.info("Skipping acoustic feature generation")
        else:
            fc = self.trainer.feature_config
        try:
            self.corpus.initialize_corpus(self.dictionary, fc)
            if self.test_transcriptions:
                self.corpus.initialize_utt_fsts()
        except CorpusError:
            if self.logger is not None:
                self.logger.warning(
                    "There was an error when initializing the corpus, likely due to missing sound files. Ignoring acoustic generation..."
                )
            self.ignore_acoustics = True

    def analyze_setup(self):
        """
        Analyzes the set up process and outputs info to the console
        """
        total_duration = sum(x.duration for x in self.corpus.files.values())
        total_duration = Decimal(str(total_duration)).quantize(Decimal("0.001"))

        ignored_count = len(self.corpus.no_transcription_files)
        ignored_count += len(self.corpus.textgrid_read_errors)
        ignored_count += len(self.corpus.decode_error_files)
        self.logger.info(
            self.corpus_analysis_template.format(
                sum(1 for x in self.corpus.files.values() if x.wav_path is not None),
                sum(1 for x in self.corpus.files.values() if x.text_type == "lab"),
                sum(1 for x in self.corpus.files.values() if x.text_type == "textgrid"),
                ignored_count,
                len(self.corpus.speakers),
                self.corpus.num_utterances,
                total_duration,
                self.analyze_oovs(),
                self.analyze_wav_errors(),
                self.analyze_missing_features(),
                self.analyze_files_with_no_transcription(),
                self.analyze_transcriptions_with_no_wavs(),
                self.analyze_textgrid_read_errors(),
                self.analyze_unreadable_text_files(),
            )
        )

    def analyze_oovs(self) -> str:
        """
        Analyzes OOVs in the corpus and constructs message

        Returns
        -------
        str
            OOV validation result message
        """
        output_dir = self.corpus.output_directory
        oov_types = self.dictionary.oovs_found
        oov_path = os.path.join(output_dir, "oovs_found.txt")
        utterance_oov_path = os.path.join(output_dir, "utterance_oovs.txt")
        if oov_types:
            total_instances = 0
            with open(utterance_oov_path, "w", encoding="utf8") as f:
                for utt, utterance in sorted(self.corpus.utterances.items()):
                    if not utterance.oovs:
                        continue
                    total_instances += len(utterance.oovs)
                    f.write(f"{utt} {', '.join(utterance.oovs)}\n")
            self.dictionary.save_oovs_found(output_dir)
            message = (
                f"There were {len(oov_types)} word types not found in the dictionary with a total of {total_instances} instances.\n\n"
                f"    Please see \n\n        {oov_path}\n\n    for a full list of the word types and \n\n        {utterance_oov_path}\n\n    for a by-utterance breakdown of "
                f"missing words."
            )
        else:
            message = (
                "There were no missing words from the dictionary. If you plan on using the a model trained "
                "on this dataset to align other datasets in the future, it is recommended that there be at "
                "least some missing words."
            )
        return message

    def analyze_wav_errors(self) -> str:
        """
        Analyzes any sound file issues in the corpus and constructs message

        Returns
        -------
        str
            Sound file validation result message
        """
        output_dir = self.corpus.output_directory
        wav_read_errors = self.corpus.sound_file_errors
        if wav_read_errors:
            path = os.path.join(output_dir, "sound_file_errors.csv")
            with open(path, "w") as f:
                for p in wav_read_errors:
                    f.write(f"{p}\n")

            message = (
                f"There were {len(wav_read_errors)} sound files that could not be read. "
                f"Please see {path} for a list."
            )
        else:
            message = "There were no sound files that could not be read."

        return message

    def analyze_missing_features(self) -> str:
        """
        Analyzes issues in feature generation in the corpus and constructs message

        Returns
        -------
        str
            Feature validation result message
        """
        if self.ignore_acoustics:
            return "Acoustic feature generation was skipped."
        output_dir = self.corpus.output_directory
        missing_features = [x for x in self.corpus.utterances.values() if x.ignored]
        if missing_features:
            path = os.path.join(output_dir, "missing_features.csv")
            with open(path, "w") as f:
                for utt in missing_features:
                    if utt.begin is not None:

                        f.write(f"{utt.file.wav_path},{utt.begin},{utt.end}\n")
                    else:
                        f.write(f"{utt.file.wav_path}\n")

            message = (
                f"There were {len(missing_features)} utterances missing features. "
                f"Please see {path} for a list."
            )
        else:
            message = "There were no utterances missing features."
        return message

    def analyze_files_with_no_transcription(self) -> str:
        """
        Analyzes issues with sound files that have no transcription files
        in the corpus and constructs message

        Returns
        -------
        str
            File matching validation result message
        """
        output_dir = self.corpus.output_directory
        if self.corpus.no_transcription_files:
            path = os.path.join(output_dir, "missing_transcriptions.csv")
            with open(path, "w") as f:
                for file_path in self.corpus.no_transcription_files:
                    f.write(f"{file_path}\n")
            message = (
                f"There were {len(self.corpus.no_transcription_files)} sound files missing transcriptions. "
                f"Please see {path} for a list."
            )
        else:
            message = "There were no sound files missing transcriptions."
        return message

    def analyze_transcriptions_with_no_wavs(self) -> str:
        """
        Analyzes issues with transcription that have no sound files
        in the corpus and constructs message

        Returns
        -------
        str
            File matching validation result message
        """
        output_dir = self.corpus.output_directory
        if self.corpus.transcriptions_without_wavs:
            path = os.path.join(output_dir, "transcriptions_missing_sound_files.csv")
            with open(path, "w") as f:
                for file_path in self.corpus.transcriptions_without_wavs:
                    f.write(f"{file_path}\n")
            message = (
                f"There were {len(self.corpus.transcriptions_without_wavs)} transcription files missing sound files. "
                f"Please see {path} for a list."
            )
        else:
            message = "There were no transcription files missing sound files."
        return message

    def analyze_textgrid_read_errors(self) -> str:
        """
        Analyzes issues with reading TextGrid files
        in the corpus and constructs message

        Returns
        -------
        str
            TextGrid validation result message
        """
        output_dir = self.corpus.output_directory
        if self.corpus.textgrid_read_errors:
            path = os.path.join(output_dir, "textgrid_read_errors.txt")
            with open(path, "w") as f:
                for k, v in self.corpus.textgrid_read_errors.items():
                    f.write(
                        f"The TextGrid file {k} gave the following error on load:\n\n{v}\n\n\n"
                    )
            message = (
                f"There were {len(self.corpus.textgrid_read_errors)} TextGrid files that could not be parsed. "
                f"Please see {path} for a list."
            )
        else:
            message = "There were no issues reading TextGrids."
        return message

    def analyze_unreadable_text_files(self) -> str:
        """
        Analyzes issues with reading text files
        in the corpus and constructs message

        Returns
        -------
        str
            Text file validation result message
        """
        output_dir = self.corpus.output_directory
        if self.corpus.decode_error_files:
            path = os.path.join(output_dir, "utf8_read_errors.csv")
            with open(path, "w") as f:
                for file_path in self.corpus.decode_error_files:
                    f.write(f"{file_path}\n")
            message = (
                f"There were {len(self.corpus.decode_error_files)} text files that could not be parsed. "
                f"Please see {path} for a list."
            )
        else:
            message = "There were no issues reading text files."
        return message

    def analyze_unaligned_utterances(self) -> None:
        """
        Analyzes issues with any unaligned files following training
        """
        unaligned_utts = self.trainer.get_unaligned_utterances()
        num_utterances = self.corpus.num_utterances
        if unaligned_utts:
            path = os.path.join(self.corpus.output_directory, "unalignable_files.csv")
            with open(path, "w") as f:
                f.write("File path,begin,end,duration,text length\n")
                for utt in unaligned_utts:
                    utterance = self.corpus.utterances[utt]
                    utt_duration = utterance.duration
                    utt_length_words = utterance.text.count(" ") + 1
                    if utterance.begin is not None:
                        f.write(
                            f"{utterance.file.wav_path},{utterance.begin},{utterance.end},{utt_duration},{utt_length_words}\n"
                        )
                    else:
                        f.write(f"{utterance.file.wav_path},,,{utt_duration},{utt_length_words}\n")
            message = (
                f"There were {len(unaligned_utts)} unalignable utterances out of {num_utterances} after the initial training. "
                f"Please see {path} for a list."
            )
        else:
            message = f"All {num_utterances} utterances were successfully aligned!"
        print(self.alignment_analysis_template.format(message))

    def validate(self):
        """
        Performs validation of the corpus
        """
        self.analyze_setup()
        if self.ignore_acoustics:
            print("Skipping test alignments.")
            return
        if not isinstance(self.trainer, PretrainedAligner):
            self.trainer.init_training(
                self.trainer.train_type, self.temp_directory, self.corpus, self.dictionary, None
            )
            self.trainer.train()
        self.trainer.align(None)
        self.analyze_unaligned_utterances()
        if self.test_transcriptions:
            self.test_utterance_transcriptions()

    def test_utterance_transcriptions(self):
        """
        Tests utterance transcriptions with simple unigram models based on the utterance text and frequent
        words in the corpus

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        self.logger.info("Checking utterance transcriptions...")

        model_directory = self.trainer.align_directory
        log_directory = os.path.join(model_directory, "log")

        try:

            jobs = [x.compile_utterance_train_graphs_arguments(self) for x in self.corpus.jobs]
            if self.trainer.feature_config.use_mp:
                run_mp(compile_utterance_train_graphs_func, jobs, log_directory)
            else:
                run_non_mp(compile_utterance_train_graphs_func, jobs, log_directory)
            self.logger.info("Utterance FSTs compiled!")
            self.logger.info("Decoding utterances (this will take some time)...")
            jobs = [x.test_utterances_arguments(self) for x in self.corpus.jobs]
            if self.trainer.feature_config.use_mp:
                run_mp(test_utterances_func, jobs, log_directory)
            else:
                run_non_mp(test_utterances_func, jobs, log_directory)
            self.logger.info("Finished decoding utterances!")

            word_mapping = self.dictionary.reversed_word_mapping
            errors = {}

            for job in jobs:
                for dict_name in job.dictionaries:
                    aligned_int = load_scp(job.out_int_paths[dict_name])
                    for utt, line in sorted(aligned_int.items()):
                        text = []
                        for t in line:
                            text.append(word_mapping[int(t)])
                        ref_text = self.corpus.utterances[utt].text.split()
                        edits = edit_distance(text, ref_text)

                        if edits:
                            errors[utt] = (edits, ref_text, text)
            if not errors:
                message = "There were no utterances with transcription issues."
            else:
                out_path = os.path.join(self.corpus.output_directory, "transcription_problems.csv")
                with open(out_path, "w") as problemf:
                    problemf.write("Utterance,Edits,Reference,Decoded\n")
                    for utt, (edits, ref_text, text) in sorted(
                        errors.items(), key=lambda x: -1 * (len(x[1][1]) + len(x[1][2]))
                    ):
                        problemf.write(f"{utt},{edits},{' '.join(ref_text)},{' '.join(text)}\n")
                message = (
                    f"There were {len(errors)} of {self.corpus.num_utterances} utterances with at least one transcription issue. "
                    f"Please see the outputted csv file {out_path}."
                )

            self.logger.info(self.transcription_analysis_template.format(message))

        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise



"""Class definitions for training G2P models"""
from __future__ import annotations

import functools
import logging
import multiprocessing as mp
import operator
import os
import queue
import random
import re
import shutil
import subprocess
import sys
import time
import traceback
from typing import TYPE_CHECKING, Any, Callable, Dict, List, NamedTuple, Optional, Set, Tuple

import tqdm

from ..config import TEMP_DIR
from ..helper import score
from ..models import G2PModel
from ..multiprocessing import Counter, Stopped
from .generator import PyniniDictionaryGenerator

try:
    import pynini
    import pywrapfst
    from pynini import Fst, TokenType
    from pywrapfst import convert

    G2P_DISABLED = False

except ImportError:
    pynini = None
    pywrapfst = None
    TokenType = None
    Fst = None

    def convert(x):
        pass

    G2P_DISABLED = True

if TYPE_CHECKING:
    from ..config.train_g2p_config import TrainG2PConfig
    from ..dictionary import Dictionary, DictionaryEntryType


Labels = List[Any]

TOKEN_TYPES = ["byte", "utf8"]
DEV_NULL = open(os.devnull, "w")
INF = float("inf")
RAND_MAX = 32767

__all__ = ["RandomStartWorker", "PairNGramAligner", "PyniniTrainer"]


class RandomStart(NamedTuple):
    """Parameters for random starts"""

    idx: int
    seed: int
    g_path: str
    p_path: str
    c_path: str
    tempdir: str
    train_opts: List[str]


def compute_validation_errors(
    gold_values: Dict[str, List[Dict[str, str]]],
    hypothesis_values: Dict[str, List[str]],
    num_jobs: int = 3,
) -> Tuple[float, float]:
    """
    Computes validation errors

    Parameters
    ----------
    gold_values: Dict
        Gold labels
    hypothesis_values: Dict
        Hypothesis labels
    num_jobs: int
        Number of jobs to use

    Returns
    -------
    float
        Word error rate
    float
        Phone error rate
    """
    # Word-level measures.
    correct = 0
    incorrect = 0
    # Label-level measures.
    total_edits = 0
    total_length = 0
    # Since the edit distance algorithm is quadratic, let's do this with
    # multiprocessing.
    with mp.Pool(num_jobs) as pool:
        to_comp = []
        for word, hyp in hypothesis_values.items():
            g = gold_values[word][0]["pronunciation"]
            hyp = [h.split(" ") for h in hyp]
            to_comp.append((g, hyp))
        gen = pool.starmap(score, to_comp)
        for (edits, length) in gen:
            if edits == 0:
                correct += 1
            else:
                incorrect += 1
            total_edits += edits
            total_length += length
        for w, gold in gold_values.items():
            if w not in hypothesis_values:
                incorrect += 1
                gold = gold[0]["pronunciation"]
                total_edits += len(gold)
                total_length += len(gold)

    return 100 * incorrect / (correct + incorrect), 100 * total_edits / total_length


class RandomStartWorker(mp.Process):
    """
    Random start worker
    """

    def __init__(
        self,
        job_q: mp.Queue,
        return_dict: Dict,
        function: Callable,
        counter: Counter,
        stopped: Stopped,
    ):
        mp.Process.__init__(self)
        self.job_q = job_q
        self.return_dict = return_dict
        self.function = function
        self.counter = counter
        self.stopped = stopped

    def run(self) -> None:
        """Run the random start worker"""
        while True:
            try:
                args = self.job_q.get(timeout=1)
            except queue.Empty:
                break
            self.job_q.task_done()
            if self.stopped.stop_check():
                continue
            try:
                fst_path, likelihood = self.function(args)
                self.return_dict[fst_path] = likelihood
            except Exception:
                self.stopped.stop()
                self.return_dict["error"] = args, Exception(
                    traceback.format_exception(*sys.exc_info())
                )
            self.counter.increment()
        return


class PairNGramAligner:
    """Produces FSA alignments for pair n-gram model training."""

    _compactor = functools.partial(convert, fst_type="compact_string")

    def __init__(self, temp_directory: str):
        self.tempdir = temp_directory
        self.g_path = os.path.join(self.tempdir, "g.far")
        self.p_path = os.path.join(self.tempdir, "p.far")
        self.c_path = os.path.join(self.tempdir, "c.fst")
        self.align_path = os.path.join(self.tempdir, "align.fst")
        self.afst_path = os.path.join(self.tempdir, "afst.far")
        self.align_log_path = os.path.join(self.tempdir, "align.log")
        self.logger = logging.getLogger("g2p_aligner")
        self.logger.setLevel(logging.DEBUG)

        handler = logging.FileHandler(self.align_log_path)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

    def align(
        self,
        # Input TSV path.
        tsv_path: str,
        # Output FAR path.
        far_path: str,
        encoder_path: str,
        # Arguments for constructing the lexicon and covering grammar.
        input_token_type: TokenType,
        input_epsilon: bool,
        output_token_type: TokenType,
        output_epsilon: bool,
        # Arguments used during the alignment phase.
        cores: int,
        random_starts: int,
        seed: int,
        batch_size: int = 0,
        delta: float = 1 / 1024,
        lr: float = 1.0,
        max_iters: int = 50,
        fst_default_cache_gc: str = "",
        fst_default_cache_gc_limit: str = "",
    ):
        """Runs the entire alignment regimen."""
        self._lexicon_covering(
            tsv_path,
            input_token_type,
            input_epsilon,
            output_token_type,
            output_epsilon,
        )
        self._alignments(
            cores,
            random_starts,
            seed,
            batch_size,
            delta,
            lr,
            max_iters,
            fst_default_cache_gc,
            fst_default_cache_gc_limit,
        )
        self._encode(far_path, encoder_path)
        self.logger.info("Success! FAR path: %s; encoder path: %s", far_path, encoder_path)

    @staticmethod
    def _label_union(labels: Set[int], epsilon: bool) -> Fst:
        """Creates FSA over a union of the labels."""
        side = pynini.Fst()
        src = side.add_state()
        side.set_start(src)
        dst = side.add_state()
        if epsilon:
            labels.add(0)
        one = pynini.Weight.one(side.weight_type())
        for label in labels:
            side.add_arc(src, pynini.Arc(label, label, one, dst))
        side.set_final(dst)
        assert side.verify(), "FST is ill-formed"
        return side

    @staticmethod
    def _narcs(f: Fst) -> int:
        """Computes the number of arcs in an FST."""
        return sum(f.num_arcs(state) for state in f.states())

    NON_SYMBOL = ("byte", "utf8")

    def _lexicon_covering(
        self,
        tsv_path: str,
        input_token_type: TokenType,
        input_epsilon: bool,
        output_token_type: TokenType,
        output_epsilon: bool,
    ) -> None:
        """Builds covering grammar and lexicon FARs."""
        # Sets of labels for the covering grammar.
        g_labels: Set[int] = set()
        p_labels: Set[int] = set()
        self.logger.info("Constructing grapheme and phoneme FARs")
        g_writer = pywrapfst.FarWriter.create(self.g_path)
        p_writer = pywrapfst.FarWriter.create(self.p_path)
        with open(tsv_path, "r") as source:
            for (linenum, line) in enumerate(source, 1):
                key = f"{linenum:08x}"
                (g, p) = line.rstrip().split("\t", 1)
                # For both G and P, we compile a FSA, store the labels, and
                # then write the compact version to the FAR.
                g_fst = pynini.accep(g, token_type=input_token_type)
                g_labels.update(g_fst.paths().ilabels())
                g_writer[key] = self._compactor(g_fst)
                p_fst = pynini.accep(p, token_type=output_token_type)
                p_labels.update(p_fst.paths().ilabels())
                p_writer[key] = self._compactor(p_fst)
        self.logger.info(f"Processed{linenum:,d} examples")
        self.logger.info("Constructing covering grammar")
        self.logger.info(f"{len(g_labels)} unique graphemes")
        g_side = self._label_union(g_labels, input_epsilon)
        self.logger.info(f"{len(p_labels)} unique phones")
        p_side = self._label_union(p_labels, output_epsilon)
        # The covering grammar is given by (G job_name P)^*.
        covering = pynini.cross(g_side, p_side).closure().optimize()
        assert covering.num_states() == 1, "Covering grammar FST is ill-formed"
        self.logger.info(
            f"Covering grammar has {PairNGramAligner._narcs(covering):,d} arcs",
        )
        covering.write(self.c_path)

    @staticmethod
    def _random_start(random_start: RandomStart) -> Tuple[str, float]:
        """Performs a single random start."""
        start = time.time()
        logger = logging.getLogger("g2p_aligner")
        # Randomize channel model.
        c_path = os.path.join(random_start.tempdir, f"c-{random_start.seed:05d}.fst")
        t_path = os.path.join(random_start.tempdir, f"t-{random_start.seed:05d}.fst")
        likelihood_path = t_path.replace(".fst", ".like")
        if not os.path.exists(t_path):
            cmd = [
                "baumwelchrandomize",
                f"--seed={random_start.seed}",
                random_start.c_path,
                c_path,
            ]
            subprocess.check_call(cmd)
            random_end = time.time()
            logger.debug(f"{random_start.seed} randomization took {random_end - start} seconds")
            # Train on randomized channel model.

            likelihood = INF
            cmd = [
                "baumwelchtrain",
                *random_start.train_opts,
                random_start.g_path,
                random_start.p_path,
                c_path,
                t_path,
            ]
            logger.debug(f"{random_start.seed} train command: {' '.join(cmd)}")
            with subprocess.Popen(cmd, stderr=subprocess.PIPE, text=True) as proc:
                # Parses STDERR to capture the likelihood.
                for line in proc.stderr:  # type: ignore
                    line = line.rstrip()
                    match = re.match(r"INFO: Iteration \d+: (-?\d*(\.\d*)?)", line)
                    assert match, line
                    likelihood = float(match.group(1))
                with open(likelihood_path, "w") as f:
                    f.write(str(likelihood))
            logger.debug(f"{random_start.seed} training took {time.time() - random_end} seconds")
        else:
            with open(likelihood_path, "r") as f:
                likelihood = f.read().strip()
        return t_path, likelihood

    def _alignments(
        self,
        cores: int,
        random_starts: int,
        seed: int,
        batch_size: Optional[int] = None,
        delta: Optional[float] = None,
        lr: Optional[float] = None,
        max_iters: Optional[int] = None,
        fst_default_cache_gc: str = "",
        fst_default_cache_gc_limit: str = "",
    ) -> None:
        """Trains the aligner and constructs the alignments FAR."""
        if not os.path.exists(self.align_path):
            self.logger.info("Training aligner")
            train_opts = []
            if batch_size:
                train_opts.append(f"--batch_size={batch_size}")
            if delta:
                train_opts.append(f"--delta={delta}")
            if fst_default_cache_gc:
                train_opts.append(f"--fst_default_cache_gc={fst_default_cache_gc}")
            if fst_default_cache_gc_limit:
                train_opts.append(f"--fst_default_cache_gc_limit={fst_default_cache_gc_limit}")
            if lr:
                train_opts.append(f"--lr={lr}")
            if max_iters:
                train_opts.append(f"--max_iters={max_iters}")
            # Constructs the actual command vectors (plus an index for logging
            # purposes).
            random.seed(seed)
            starts = [
                (
                    RandomStart(
                        idx,
                        seed,
                        self.g_path,
                        self.p_path,
                        self.c_path,
                        self.tempdir,
                        train_opts,
                    )
                )
                for (idx, seed) in enumerate(random.sample(range(1, RAND_MAX), random_starts), 1)
            ]
            stopped = Stopped()
            num_commands = len(starts)
            if cores > len(starts):
                cores = len(starts)
            job_queue = mp.JoinableQueue(cores + 2)

            # Actually runs starts.
            self.logger.info("Random starts")
            print("Calculating alignments...")
            begin = time.time()
            last_value = 0
            ind = 0
            with tqdm.tqdm(total=num_commands) as pbar:
                while True:
                    if ind == num_commands:
                        break
                    try:
                        job_queue.put(starts[ind], False)
                    except queue.Full:
                        break
                    ind += 1
                manager = mp.Manager()
                return_dict = manager.dict()
                procs = []
                counter = Counter()
                for _ in range(cores):
                    p = RandomStartWorker(
                        job_queue, return_dict, self._random_start, counter, stopped
                    )
                    procs.append(p)
                    p.start()
                while True:
                    if ind == num_commands:
                        break
                    job_queue.put(starts[ind])
                    value = counter.value()
                    pbar.update(value - last_value)
                    last_value = value
                    ind += 1
                while True:
                    time.sleep(30)
                    value = counter.value()
                    if value != last_value:
                        pbar.update(value - last_value)
                        last_value = value
                    if value >= random_starts:
                        break
                job_queue.join()
                for p in procs:
                    p.join()
            if "error" in return_dict:
                element, exc = return_dict["error"]
                print(element)
                raise exc
            (best_fst, best_likelihood) = min(return_dict.items(), key=operator.itemgetter(1))
            self.logger.info(f"Best likelihood: {best_likelihood}")
            self.logger.debug(
                f"Ran {random_starts} random starts in {time.time() - begin} seconds"
            )
            # Moves best likelihood solution to the requested location.
            shutil.move(best_fst, self.align_path)
        self.logger.info("Computing alignments")
        cmd = ["baumwelchdecode"]
        if fst_default_cache_gc:
            cmd.append(f"--fst_default_cache_gc={fst_default_cache_gc}")
        if fst_default_cache_gc_limit:
            cmd.append(f"--fst_default_cache_gc_limit={fst_default_cache_gc_limit}")
        cmd.append(self.g_path)
        cmd.append(self.p_path)
        cmd.append(self.align_path)
        cmd.append(self.afst_path)
        self.logger.debug(f"Subprocess call: {cmd}")
        subprocess.check_call(cmd)

    def _encode(self, far_path: str, encoder_path: str) -> None:
        """Encodes the alignments."""
        self.logger.info("Encoding the alignments as FSAs")
        encoder = pywrapfst.EncodeMapper(encode_labels=True)
        a_reader = pywrapfst.FarReader.open(self.afst_path)
        a_writer = pywrapfst.FarWriter.create(far_path)
        # Curries converter function for the FAR.
        converter = functools.partial(pywrapfst.convert, fst_type="vector")
        while not a_reader.done():
            key = a_reader.get_key()
            fst = converter(a_reader.get_fst())
            fst.encode(encoder)
            a_writer[key] = self._compactor(fst)
            try:
                next(a_reader)
            except StopIteration:
                break
        encoder.write(encoder_path)


class PyniniTrainer:
    """
    Class for G2P trainer that uses Pynini functionality

    Parameters
    ----------
    dictionary: Dictionary
        Dictionary to train from`
    model_path: str
        Output model path
    train_config: TrainG2PConfig
        Configuration for training G2P model
    temp_directory: str, optional
        Temporary directory, defaults to MFA's temporary directory
    input_epsilon: bool
        Flag for whether to allow for epsilon on input strings, default True
    output_epsilon: bool
        Flag for whether to allow for epsilon on output strings, default True
    num_jobs: int
        Number processes to use
    verbose: bool
        Flag for provide debug output to the terminal
    """

    def __init__(
        self,
        dictionary: Dictionary,
        model_path: str,
        train_config: TrainG2PConfig,
        temp_directory: Optional[str] = None,
        input_epsilon: bool = True,
        output_epsilon: bool = True,
        num_jobs: int = 3,
        verbose: bool = False,
    ):
        super(PyniniTrainer, self).__init__()
        if not temp_directory:
            temp_directory = TEMP_DIR
        self.temp_directory = os.path.join(temp_directory, "G2P")
        self.train_config = train_config
        self.verbose = verbose
        self.models_temp_dir = os.path.join(temp_directory, "models", "G2P")

        self.name, _ = os.path.splitext(os.path.basename(model_path))
        self.temp_directory = os.path.join(self.temp_directory, self.name)
        os.makedirs(self.temp_directory, exist_ok=True)
        os.makedirs(self.models_temp_dir, exist_ok=True)
        self.model_path = model_path
        self.fst_path = os.path.join(self.temp_directory, "model.fst")
        self.far_path = os.path.join(self.temp_directory, self.name + ".far")
        self.encoder_path = os.path.join(self.temp_directory, self.name + ".enc")
        self.dictionary = dictionary
        self.input_epsilon = input_epsilon
        self.output_epsilon = output_epsilon
        self.num_jobs = num_jobs
        if not self.train_config.use_mp:
            self.num_jobs = 1
        self.fst_default_cache_gc = ""
        self.fst_default_cache_gc_limit = ""
        self.train_log_path = os.path.join(self.temp_directory, "train.log")

        self.logger = logging.getLogger("g2p_trainer")
        self.logger.setLevel(logging.DEBUG)

        handler = logging.FileHandler(self.train_log_path)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        handler = logging.StreamHandler(sys.stdout)
        if self.verbose:
            handler.setLevel(logging.DEBUG)
        else:
            handler.setLevel(logging.INFO)
        formatter = logging.Formatter("%(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.model_log_path = os.path.join(self.temp_directory, "model.log")
        self.sym_path = os.path.join(self.temp_directory, "phones.sym")
        self.output_token_type = None

    def clean_up(self) -> None:
        """
        Clean up temporary files
        """
        for name in os.listdir(self.temp_directory):
            path = os.path.join(self.temp_directory, name)
            if os.path.isdir(path):
                shutil.rmtree(path, ignore_errors=True)
            elif not name.endswith(".log"):
                os.remove(path)

    def generate_model(self) -> None:
        """
        Generate an ngram G2P model from FAR strings
        """
        assert os.path.exists(self.far_path)
        with open(self.model_log_path, "w", encoding="utf8") as logf:
            ngram_count_path = os.path.join(self.temp_directory, "ngram.count")
            ngram_make_path = os.path.join(self.temp_directory, "ngram.make")
            ngram_shrink_path = os.path.join(self.temp_directory, "ngram.shrink")
            ngramcount_proc = subprocess.Popen(
                [
                    "ngramcount",
                    "--require_symbols=false",
                    "--order={}".format(self.train_config.order),
                    self.far_path,
                    ngram_count_path,
                ],
                stderr=logf,
            )
            ngramcount_proc.communicate()

            ngrammake_proc = subprocess.Popen(
                [
                    "ngrammake",
                    "--method=" + self.train_config.smoothing_method,
                    ngram_count_path,
                    ngram_make_path,
                ],
                stderr=logf,
            )
            ngrammake_proc.communicate()

            ngramshrink_proc = subprocess.Popen(
                [
                    "ngramshrink",
                    "--method=" + self.train_config.pruning_method,
                    "--target_number_of_ngrams={}".format(self.train_config.model_size),
                    ngram_make_path,
                    ngram_shrink_path,
                ],
                stderr=logf,
            )
            ngramshrink_proc.communicate()

            fstencode_proc = subprocess.Popen(
                ["fstencode", "--decode", ngram_shrink_path, self.encoder_path, self.fst_path],
                stderr=logf,
            )
            fstencode_proc.communicate()

        os.remove(ngram_count_path)
        os.remove(ngram_make_path)
        os.remove(ngram_shrink_path)

        directory, filename = os.path.split(self.model_path)
        basename, _ = os.path.splitext(filename)
        model = G2PModel.empty(basename, root_directory=self.models_temp_dir)
        model.add_meta_file(self.dictionary)
        model.add_fst_model(self.temp_directory)
        model.add_sym_path(self.temp_directory)
        if directory:
            os.makedirs(directory, exist_ok=True)
        basename, _ = os.path.splitext(self.model_path)
        model.dump(basename)
        model.clean_up()
        self.clean_up()
        self.logger.info(f"Saved model to {self.model_path}")

    def train(self, word_dict: Optional[Dict[str, DictionaryEntryType]] = None) -> None:
        """
        Train a G2P model

        Parameters
        ----------
        word_dict: Dict[str, DictionaryEntryType]
            Dictionary of words to pronunciations, optional, defaults to the dictionary's
            set of words
        """
        input_path = os.path.join(self.temp_directory, "input.txt")
        phones_path = os.path.join(self.temp_directory, "phones_only.txt")
        if word_dict is None:
            word_dict = self.dictionary.actual_words
        with open(input_path, "w", encoding="utf8") as f2, open(
            phones_path, "w", encoding="utf8"
        ) as phonef:
            for word, v in word_dict.items():
                if re.match(r"\W", word) is not None:
                    continue
                for v2 in v:
                    f2.write(f"{word}\t{' '.join(v2['pronunciation'])}\n")
                for v2 in v:
                    phonef.write(f"{' '.join(v2['pronunciation'])}\n")
        subprocess.call(["ngramsymbols", phones_path, self.sym_path])
        os.remove(phones_path)
        aligner = PairNGramAligner(self.temp_directory)
        input_token_type = "utf8"
        self.output_token_type = pynini.SymbolTable.read_text(self.sym_path)
        begin = time.time()
        if not os.path.exists(self.far_path) or not os.path.exists(self.encoder_path):
            aligner.align(
                input_path,
                self.far_path,
                self.encoder_path,
                input_token_type,
                self.input_epsilon,
                self.output_token_type,
                self.output_epsilon,
                self.num_jobs,
                self.train_config.random_starts,
                self.train_config.seed,
                self.train_config.batch_size,
                self.train_config.delta,
                self.train_config.lr,
                self.train_config.max_iterations,
                self.fst_default_cache_gc,
                self.fst_default_cache_gc_limit,
            )
        self.logger.debug(f"Aligning {len(word_dict)} words took {time.time() - begin} seconds")
        begin = time.time()
        self.generate_model()
        self.logger.debug(
            f"Generating model for {len(word_dict)} words took {time.time() - begin} seconds"
        )

    def validate(self) -> None:
        """
        Validate the G2P model against held out data
        """

        word_dict = self.dictionary.actual_words
        validation = 0.1
        words = word_dict.keys()
        total_items = len(words)
        validation_items = int(total_items * validation)
        validation_words = random.sample(words, validation_items)
        training_dictionary = {k: v for k, v in word_dict.items() if k not in validation_words}
        validation_dictionary = {k: v for k, v in word_dict.items() if k in validation_words}
        train_graphemes = set()
        for k in word_dict.keys():
            train_graphemes.update(k)
        self.train(training_dictionary)

        model = G2PModel(self.model_path, root_directory=self.temp_directory)

        gen = PyniniDictionaryGenerator(
            model,
            validation_dictionary.keys(),
            temp_directory=os.path.join(self.temp_directory, "validation"),
            num_jobs=self.num_jobs,
            num_pronunciations=self.train_config.num_pronunciations,
        )
        output = gen.generate()
        begin = time.time()
        wer, ler = compute_validation_errors(validation_dictionary, output, num_jobs=self.num_jobs)
        print(f"WER:\t{wer:.2f}")
        print(f"LER:\t{ler:.2f}")
        self.logger.info(f"WER:\t{wer:.2f}")
        self.logger.info(f"LER:\t{ler:.2f}")
        self.logger.debug(
            f"Computation of errors for {len(validation_dictionary)} words took {time.time() - begin} seconds"
        )



"""Classes for configuring feature generation"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Text, Union

from ..exceptions import ConfigError
from .base_config import BaseConfig

if TYPE_CHECKING:
    SpeakerCharacterType = Union[str, int]

__all__ = ["make_safe", "FeatureConfig"]


def make_safe(value: Any) -> str:
    """
    Transform an arbitrary value into a string

    Parameters
    ----------
    value: Any
        Value to make safe

    Returns
    -------
    str
        Safe value
    """
    if isinstance(value, bool):
        return str(value).lower()
    return str(value)


class FeatureConfig(BaseConfig):
    """
    Class to store configuration information about MFCC generation

    Attributes
    ----------
    directory : str
        Path of the directory to store outputs
    type : str
        Feature type, defaults to "mfcc"
    deltas : bool
        Flag for whether deltas from previous frames are included in the features, defaults to True
    lda : bool
        Flag for whether LDA is run on the features, requires an lda.mat to generate, defaults to False
    fmllr : bool
        Flag for whether speaker adaptation should be run, defaults to False
    use_energy : bool
        Flag for whether first coefficient should be used, defaults to False
    frame_shift : int
        number of milliseconds between frames, defaults to 10
    snip_edges : bool
        Flag for enabling Kaldi's snip edges, should be better time precision
    pitch : bool
        Flag for including pitch in features, currently nonfunctional, defaults to False
    low_frequency : int
        Frequency floor
    high_frequency : int
        Frequency ceiling
    sample_frequency : int
        Sampling frequency
    allow_downsample : bool
        Flag for whether to allow downsampling, default is True
    allow_upsample : bool
        Flag for whether to allow upsampling, default is True
    splice_left_context : int or None
        Number of frames to splice on the left for calculating LDA
    splice_right_context : int or None
        Number of frames to splice on the right for calculating LDA
    use_mp : bool
        Flag for using multiprocessing, defaults to True
    """

    deprecated_flags = {"lda", "deltas"}

    def __init__(self):
        self.type = "mfcc"
        self.deltas = True
        self.fmllr = False
        self.lda = False
        self.use_energy = False
        self.frame_shift = 10
        self.snip_edges = True
        self.pitch = False
        self.low_frequency = 20
        self.high_frequency = 7800
        self.sample_frequency = 16000
        self.allow_downsample = True
        self.allow_upsample = True
        self.splice_left_context = 3
        self.splice_right_context = 3
        self.use_mp = True

    def params(self) -> Dict[Text, Any]:
        """Parameters for feature generation"""
        return {
            "type": self.type,
            "use_energy": self.use_energy,
            "frame_shift": self.frame_shift,
            "snip_edges": self.snip_edges,
            "low_frequency": self.low_frequency,
            "high_frequency": self.high_frequency,
            "sample_frequency": self.sample_frequency,
            "allow_downsample": self.allow_downsample,
            "allow_upsample": self.allow_upsample,
            "pitch": self.pitch,
            "fmllr": self.fmllr,
            "splice_left_context": self.splice_left_context,
            "splice_right_context": self.splice_right_context,
        }

    @property
    def mfcc_options(self) -> Dict[Text, Any]:
        """Parameters to use in computing MFCC features."""
        return {
            "use-energy": self.use_energy,
            "frame-shift": self.frame_shift,
            "low-freq": self.low_frequency,
            "high-freq": self.high_frequency,
            "sample-frequency": self.sample_frequency,
            "allow-downsample": self.allow_downsample,
            "allow-upsample": self.allow_upsample,
            "snip-edges": self.snip_edges,
        }

    def update(self, data: Dict[str, Any]) -> None:
        """
        Update configuration with new data

        Parameters
        ----------
        data: Dict[str, Any]
            New data
        """
        for k, v in data.items():
            if k in self.deprecated_flags:
                continue
            if not hasattr(self, k):
                raise ConfigError("No field found for key {}".format(k))
            setattr(self, k, v)

    @property
    def feature_id(self) -> str:
        """Deprecated feature ID"""
        return "feats"
